/*
 * kb.c --
 *
 *  Last Modified: Fri Sep 22 16:43:53 1995 by Bryan M. Kramer
 *
 * Copyright (c) 1990-1994 University of Toronto.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF Toronto BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * Toronto HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF Toronto SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF Toronto HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#include <stdio.h>
#include <tcl.h>
#include <tk.h>
#include <string.h>

# include "kb.h"


extern int tds_connection_exists;


KBObject* proposition_class;
KBObject* token_class;
KBObject* attribute_class;
KBObject* class_class;
KBObject* m1class_class;
KBObject* m2class_class;
KBObject* sclass_class;
KBObject* m3class_class;
KBObject* m4class_class;

KBObject* number_class;
KBObject* string_class;
KBObject* pstring_class;

Boolean initializing;

/* WINGS BEGIN */
#ifdef WINGS_UNIX
Tcl_Interp* confirm_interp;
#endif
/* WINGS END */

Source* current_source;
Source* default_source;

static int unique_integer = 1;

int ckb_connection_exists = 0;

# ifdef CKB
extern int ckb_download ();
# endif

char* source_names[] = {"none", "file", "tds", "tcl", "builtin", "ckb"};

void rep_wait ();

int debug_conn;

extern Tcl_Interp* global_interp;
static int rep_error_code;
static Tcl_DString rep_result;

KBList* cons (obj, list)
    KBObject* obj;
    KBList* list;
{
    KBList* cell;

    cell = new (KBList);
    cell->object = obj;
    cell->next = list;

    return cell;
}


KBList* list_find (obj, list)
    register KBObject* obj;
    KBList* list;
{
    register KBList* cell;

    for (cell = list; cell; cell = cell->next)
    {
	if (cell->object == obj)
	    return cell;
    }

    return nil;
}


void list_print (list, stream)
    KBList* list;
    FILE* stream;
{
  register KBList* cell;
  register Boolean first = true;

  for (cell = list; cell; cell = cell->next)
    {
      if (first)
	{
	  fprintf (stream, "%s", cell->object->name);
	  first = false;
	}
      else
	fprintf (stream, ", %s", cell->object->name);
    }
}


int list_length (list)
    KBList* list;
{
    int n;

    n = 0;
    for (; list; list = list->next)
	n++;

    return n;
}

Boolean list_equal (list1, list2)
    KBList* list1;
    KBList* list2;
{
    KBList* ptr;
    Boolean result;
    Boolean found;

    if (list_length (list1) != list_length (list2))
	return false;

    result = true;
    for (; result && list1; list1 = list1->next)
    {
	found = false;
	for (ptr = list2; ! found && ptr; ptr = ptr->next)
	{
	    if (ptr->object == list1->object)
		found = true;
	}
	if (! found)
	    result = false;
    }
    return result;
}


void free_list (list)
    KBList* list;
{
    KBList* ptr;
    KBList* next;

/* WINGS BEGIN */
/*    for (ptr = list; ptr; ptr = next) */
/*    { */
/*	next = ptr->next; */
/*	free (ptr); */
/*    } */
/* WINGS END */
}


# define HASH '#'
# define INIT_ALLOCATION 3000
# define OBJ_REALLOC_SIZE 1000

static KBObject** all_objects;
static int obj_id;
static int dump_start;
static int num_allocated_objects;

void reallocate_object_vector ()
{
    KBObject** newv;
    int i;
    int new_size;

    new_size = num_allocated_objects + OBJ_REALLOC_SIZE;
    newv = (KBObject**) malloc (new_size * sizeof (KBObject*));
    for (i = 0; i <  num_allocated_objects; i++)
	newv[i] = all_objects[i];

/* WINGS BEGIN */
/*    free (all_objects); */
/* WINGS END */
    all_objects = newv;
    num_allocated_objects = new_size;
}

Tcl_HashTable kb_symbols;

/* interpret the name #n as an index except if n > number of existing objects */

KBObject* new_object (name, attr_of)
    char* name;
    KBObject* attr_of;
{
    KBObject* obj;
    Tcl_HashEntry* entry;
    int newp;
    int id;
    char buf[1000];

    if (name && *name == HASH)
    {
	id = atoi (name + 1);
	if (id > 0 && id < obj_id)
	    return *(all_objects + id);
    }

    entry = nil;

    if (name == nil || attr_of)
	newp = 1;
    else
	entry = Tcl_CreateHashEntry (&kb_symbols, name, &newp);

    if (newp)
    {
	while (obj_id >= num_allocated_objects)
	    reallocate_object_vector ();

	obj = new (KBObject);
	all_objects[obj_id] = obj;
	sprintf (buf, "#%d", obj_id);
	obj_id++;
	obj->id = strdup (buf);
	obj->seen = -1;
	obj->name = name ? strdup (name) : nil;
	obj->inst_of = nil;
	obj->isa = nil;
	obj->children = nil;
	obj->instances = nil;
	obj->attributes = nil;
	obj->attribute_of = attr_of;
	obj->back_ptr = nil;
	obj->value = nil;
	obj->status = attr_of ? ATTRIBUTE : UNDEFINED;
	obj->source = current_source;
	obj->client_data = 0;

	if (entry)
	    Tcl_SetHashValue (entry, obj);
    }
    else
	obj = (KBObject*) Tcl_GetHashValue (entry);

    return obj;
}

KBObject* ith_object (i)
    int i;
{
    if (i > 0 && i < obj_id)
	return *(all_objects + i);
    else
	return nil;
}


void* client_data (obj)
    KBObject* obj;
{
    return obj->client_data;
}


void set_client_data (obj, val)
    KBObject* obj;
    void* val;
{
    obj->client_data = val;
}

KBObject* get_object (name)
    char* name;
{
    KBObject* obj;
    Tcl_HashEntry* entry;
    int id;

    if (! name)
	return nil;

    if (*name == HASH)
    {
	id = atoi (name + 1);
	if (id > 0 && id < obj_id)
	    return *(all_objects + id);
    }

    entry = Tcl_FindHashEntry (&kb_symbols, name);
    if (entry)
	obj = (KBObject*) Tcl_GetHashValue (entry);
    else
	obj = nil;

    return obj;
}

/* change name of object and rehash it */
/* WARNING - I'm not checking if new name clobbers another object !!! */

void rename_obj (obj, new_name)
    KBObject* obj;
    char* new_name;
{
    Tcl_HashEntry* entry;
    int id;
    int newp;
    char* name = obj->name;

    if (obj->attribute_of)
    {
	/* for attributes, do none of the rehashing below */
	obj->name = strdup (new_name);
    }
    else
    {
	if (name)
	{
	    /* deal with old name */

	    if (*name != HASH)
	    {
		entry = Tcl_FindHashEntry (&kb_symbols, name);
		if (entry)
		{
		    Tcl_DeleteHashEntry (entry);
		}
	    }
	}
	obj->name = strdup (new_name);
	if (new_name && *new_name)
	{
	    entry = Tcl_CreateHashEntry (&kb_symbols, new_name, &newp);
	    if (entry)
		Tcl_SetHashValue (entry, obj);
	}
    }
/* WINGS BEGIN */
/*    if (name) */
/*	free (name); */
/* WINGS END */
}


/* static stuff to keep track of what parser is defining */
/* mark the object defined by turning off undefined flag */

static KBList* parser_new_objects;

void record_new_object (obj)
    KBObject* obj;
{
    push (obj, parser_new_objects);
    obj->status &= ~UNDEFINED;
}

void reset_new_object_list ()
{
    free_list (parser_new_objects);
    parser_new_objects = nil;
}

void set_inst_received (name)
    char* name;
{
    KBObject* obj = new_object (name, 0);
    obj->status |= TDS_INST_RECVD;
}

int check_inst_received (name)
    char* name;
{
    KBObject* obj = new_object (name, 0);
    return (int) (obj->status & TDS_INST_RECVD);
}

void set_children_received (name)
    char* name;
{
    KBObject* obj = new_object (name, 0);
    obj->status |= TDS_CHILDREN_RECVD;
}

int check_children_received (name)
    char* name;
{
    KBObject* obj = new_object (name, 0);
    return (int) (obj->status & TDS_CHILDREN_RECVD);
}


# define CONFIRM(clientData,interp,name,routine)  { if (tds_connection_exists) { confirm_object(clientData,interp,name,routine); } else if (ckb_connection_exists) {ckb_confirm_object (clientData,interp,name,routine); }}
# define CONFIRM_CHILDREN(clientData,interp,name)  { if (tds_connection_exists) { confirm_children(clientData,interp,name);  } else if (ckb_connection_exists) ckb_confirm_children (clientData,interp,name);}
# define CONFIRM_INSTANCES(clientData,interp,name)  { if (tds_connection_exists) { confirm_instances(clientData,interp,name);}  else if (ckb_connection_exists) ckb_confirm_instances (clientData,interp,name); }
# define CONFIRM_CHILDREN_O(clientData,interp,obj)  { if (tds_connection_exists) { confirm_children_o(clientData,interp,obj);  } else if (ckb_connection_exists) ckb_confirm_children_o (clientData,interp,obj); }
# define CONFIRM_INSTANCES_O(clientData,interp,obj)  { if (tds_connection_exists) { confirm_instances_o(clientData,interp,obj);}  else if (ckb_connection_exists) ckb_confirm_instances_o (clientData,interp,obj); }


/* code to interact with TDS when connected */

void confirm_object (clientData, interp, name, routine)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
    char* routine;
{
    int i;
    KBObject* obj;

    if (*name == HASH)
	obj = get_object (name);
    else
	obj = new_object (name, nil);
    if (obj)
    {
	if (STATUS (obj, UNDEFINED) && obj->name)
	{
	    if (STATUS (obj, TDS_REQUESTED))
	    {
		if (! STATUS (obj, TDS_UNKNOWN))
		    rep_wait (clientData, interp);
	    }
	    else
	    {
		if (debug_conn)
		    fprintf (stderr, "Requesting (%s) %s %s %x\n",
			     routine ? routine : "",
			     name, obj->name ? obj->name : "<>", obj->status);
		obj->status |= TDS_REQUESTED;
		RepRequestObject (clientData, interp, obj->name, nil);
	    }
	}
    }
}



void confirm_children_o (clientData, interp, obj)
    ClientData clientData;
    Tcl_Interp* interp;
    KBObject* obj;
{
    int i;

    if (obj)
    {
	if (! STATUS (obj, TDS_CHILDREN_RECVD) && obj->name)
	{
	    /* this is false until we implement error routine */
	    if (false && STATUS (obj, TDS_CHILDREN_REQ))
	    {
		rep_wait (clientData, interp);
	    }
	    else
	    {
		if (debug_conn)
		    fprintf (stderr, "Requesting children %s %x\n", obj->name, obj->status);
		obj->status |= TDS_CHILDREN_REQ;
		RepAsk (clientData, interp, "getChildrenNames", obj->name, nil);
	    }
	}
    }
}

void confirm_children (clientData, interp, name)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
{
    int i;
    KBObject* obj;

    if (*name == HASH)
	obj = get_object (name);
    else
	obj = new_object (name, nil);

    if (obj)
	confirm_children_o (clientData, interp, obj);
}



void confirm_instances_o (clientData, interp, obj)
    ClientData clientData;
    Tcl_Interp* interp;
    KBObject* obj;
{
    int i;

    if (obj)
    {
	if (! STATUS (obj, TDS_INST_RECVD) && obj->name)
	{
	    /* this is false until we implement error routine */
	    if (false && STATUS (obj, TDS_INST_REQ))
	    {
		rep_wait (clientData, interp);
	    }
	    else
	    {
		if (debug_conn)
		    fprintf (stderr, "Requesting instances %s %x\n", obj->name, obj->status);
		obj->status |= TDS_INST_REQ;
		RepAsk (clientData, interp, "getInstanceNames", obj->name, nil);
	    }
	}
    }
}

void confirm_instances (clientData, interp, name)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
{
    int i;
    KBObject* obj;

    if (*name == HASH)
	obj = get_object (name);
    else
	obj = new_object (name, nil);

    if (obj)
	confirm_instances_o (clientData, interp, obj);
}




/* code to interact with ckb when connected */

KBObject* ckb_confirm_object (clientData, interp, name, routine)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
    char* routine;
{
    int i;
    KBObject* obj;

    obj = get_object (name);
    if (obj)
    {
	if (STATUS (obj, UNDEFINED) && obj->source->source_type == SOURCE_CKB && obj->name
	    && obj->client_data)
	{
# ifdef CKB
	    ckb_download_inst (confirm_interp, routine, obj->name, obj->client_data);
	    Tcl_ResetResult (confirm_interp);
# endif
	}
    }
    else
    {
	/* need to select on name; not implemented yet */
    }
    return obj;
}

KBObject* ckb_confirm_instances_o (clientData, interp, obj)
    ClientData clientData;
    Tcl_Interp* interp;
    KBObject* obj;
{
    int i;

    if (obj)
    {
	if (! STATUS (obj, TDS_INST_RECVD) && obj->name && obj->source->source_type == SOURCE_CKB)
	{
# ifdef CKB
	    ckb_download (obj->name, 0, confirm_interp, 0, 1, 1);
	    Tcl_ResetResult (confirm_interp);
# endif
	}
    }
    return obj;
}

KBObject* ckb_confirm_instances (clientData, interp, name)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
{
    int i;
    KBObject* obj;


    obj = get_object (name);

    if (obj)
	return ckb_confirm_instances_o (clientData, interp, obj);
    else
	return obj;
}



KBObject* ckb_confirm_children_o (clientData, interp, obj)
    ClientData clientData;
    Tcl_Interp* interp;
    KBObject* obj;
{
    int i;

    if (obj)
    {
	if (! STATUS (obj, TDS_CHILDREN_RECVD) && obj->name
	    && obj->source->source_type == SOURCE_CKB)
	{
# ifdef CKB
	    ckb_download (obj->name, 0, confirm_interp, 0, 2, 0);
	    Tcl_ResetResult (confirm_interp);
# endif
	}
    }
    return obj;
}

KBObject* ckb_confirm_children (clientData, interp, name)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
{
    int i;
    KBObject* obj;

    obj = get_object (name);

    if (obj)
	ckb_confirm_children_o (clientData, interp, obj);

    return obj;
}

/* is object 1 a subclass of object 2 */

Boolean subclass_of1 (obj1, obj2, seen)
    KBObject* obj1;
    KBObject* obj2;
    int seen;
{
    KBList* ptr;

    if (seen == obj1->seen)
	return false;

    obj1->seen = seen;

    if (obj2 == proposition_class)
	return true;

    if (obj1 == obj2)
	return true;
    else
    {
	for (ptr = obj1->isa; ptr; ptr = ptr->next)
	{
	    if (subclass_of1 (ptr->object, obj2, seen))
		return true;
	}
    }
    return false;
}


Boolean subclass_of (obj1, obj2, depth)
    KBObject* obj1;
    KBObject* obj2;
{
  unique_integer++;
  return subclass_of1 (obj1, obj2, unique_integer);
}

Boolean is_instance_of (obj, class)
    KBObject* obj;
    KBObject* class;
{
    KBList* ptr;

    for (ptr = obj->inst_of; ptr; ptr = ptr->next)
    {
	if (subclass_of (ptr->object, class, 0))
	    return true;
    }
    return false;
}


/* assume each list is already minimal */

KBList* intersect_type_lists (list1, list2)
    KBList* list1;
    KBList* list2;
{
    KBList* result;
    KBList* ptr;
    Boolean hit;

    result = nil;

    for (; list1; list1 = list1->next)
    {
	hit = false;
	for (ptr = list2; ! hit && ptr; ptr = ptr->next)
	{
	    /* take equal ones from first list */
	    if (ptr->object != list1->object && subclass_of (ptr->object, list1->object, 0))
		hit = true;
	}
	if (! hit)
	    push (list1->object, result);
    }

    for (; list2; list2 = list2->next)
    {
	hit = false;
	for (ptr = list1; ! hit && ptr; ptr = ptr->next)
	{
	    if (subclass_of (ptr->object, list2->object, 0))
		hit = true;
	}
	if (! hit)
	    push (list2->object, result);
    }

    return result;
}



KBList* minimize_classes (classes, do_free)
    KBList* classes;
    Boolean do_free;
{
    KBList* ptr;
    KBList* op;
    KBList* result;
    Boolean hit;

    result = nil;
    for (ptr = classes; ptr; ptr = ptr->next)
    {
	hit = false;
	for (op = classes; op; op = op->next)
	{
	    if (op != ptr && subclass_of (op->object, ptr->object, 0))
	    {
		hit = true;
		break;
	    }
	}
	if (! hit)
	{
	    result = cons (ptr->object, result);
	}
    }
    if (do_free)
      free_list (classes);

    return result;
}


KBList* maximize_classes (classes)
    KBList* classes;
{
    KBList* ptr;
    KBList* op;
    KBList* result;
    Boolean hit;

    result = nil;
    for (ptr = classes; ptr; ptr = ptr->next)
    {
	hit = false;
	for (op = classes; op; op = op->next)
	{
	    if (op != ptr && subclass_of (ptr->object, op->object, 0))
	    {
		hit = true;
		break;
	    }
	}
	if (! hit)
	    result = cons (ptr->object, result);
    }
    free_list (classes);

    return result;
}

KBList* list_remove (obj, list)
    KBObject* obj;
    KBList * list;
{
    KBList* ptr;
    KBList* next;
    KBList* prev;
    KBList* result;

    result = nil;
    prev = nil;
    for (ptr = list; ptr; ptr = next)
    {
	next = ptr->next;
	if (ptr->object == obj)
	{
	    if (prev)
		prev->next = next;
	    else
		result = next;
/* WINGS BEGIN */
/*	    free (ptr); */
/* WINGS END */
	}
	else
	{
	    prev = ptr;
	    if (! result)
		result = ptr;
	}
    }
    return result;
}


void add_isa_parent (obj, parent, newp)
    KBObject* obj;
    KBObject* parent;
    Boolean newp;
{
    Boolean added;
    KBList* parents;
    KBList* next;
    KBList* prev;
    KBList* result;
    Boolean hit;

    /* iterate through existing parents and minimize */
    next = nil;
    added = true;
    result = nil;
    prev = nil;
    for (parents = obj->isa; parents; parents = next)
    {
	next = parents->next;
	if (subclass_of (parents->object, parent, 0))
	{
	    /* specialized by existing parent */
	    added = false;
	    prev = parents;
	    if (! result)
		result = parents;
	}
	else if (subclass_of (parent, parents->object, 0))
	{
	    /* specializes existing parent 1: remove parent */
	    remove (obj, parents->object->children);
	    if (prev)
		prev->next = next;
	    else
		result = next;

/* WINGS BEGIN */
/*	    free (parents); */
/* WINGS END */
	}
	else
	{
	    prev = parents;
	    if (! result)
		result = parents;
	}
    }

    if (added)
    {
	obj->isa = cons (parent, result);
	if (newp)
	    pushf (obj, parent->children);
	else
	    push (obj, parent->children);
    }
    else
	obj->isa = result;
}

/* make sure that is subclass of the right toplevel class */

void verify_isa (obj, newp)
    KBObject* obj;
    Boolean   newp;
{
    if (initializing)
	return;
    if (is_instance_of (obj, token_class))
	;
    else if (is_instance_of (obj, sclass_class))
	add_isa_parent (obj, token_class, newp);
    else if (is_instance_of (obj, m1class_class))
	add_isa_parent (obj, sclass_class, newp);
    else if (is_instance_of (obj, m2class_class))
	add_isa_parent (obj, m1class_class, newp);
    else if (is_instance_of (obj, m3class_class))
	add_isa_parent (obj, m2class_class, newp);
    else if (is_instance_of (obj, m4class_class))
	add_isa_parent (obj, m3class_class, newp);
    else if (is_instance_of (obj, class_class))
	add_isa_parent (obj, class_class, newp);
}



void add_instance_of (obj, parent, newobj)
    KBObject* obj;
    KBObject* parent;
    Boolean newobj;
{
    Boolean added;
    KBList* parents;
    KBList* next;
    KBList* prev;
    KBList* result;
    Boolean hit;

    /* iterate through existing parents and minimize */
    next = nil;
    added = true;
    result = nil;
    prev = nil;
    for (parents = obj->inst_of; parents; parents = next)
    {
	next = parents->next;
	if (subclass_of (parents->object, parent, 0))
	{
	    /* specialized by existing parent */
	    added = false;
	    prev = parents;
	    if (! result)
		result = parents;
	}
	else if (subclass_of (parent, parents->object, 0))
	{
	    /* specializes existing parent 1: remove parent */
	    remove (obj, parents->object->instances);
	    if (prev)
		prev->next = next;
	    else
		result = next;

/* WINGS BEGIN */
/*	    free (parents); */
/* WINGS END */
	}
	else
	{
	    prev = parents;
	    if (! result)
		result = parents;
	}
    }
    if (prev)
	prev->next = nil;
    if (added)
    {
	obj->inst_of = cons (parent, result);
	if (newobj)
	    pushf (obj, parent->instances);
	else
	    push (obj, parent->instances);
    }
    else
	obj->inst_of = result;

    verify_isa (obj, newobj);
}


void add_instance_of_list (obj, classes, newobj)
    KBObject* obj;
    KBList* classes;
    Boolean newobj;
{
    for (; classes; classes = classes->next)
    {
	add_instance_of (obj, classes->object, newobj);
	newobj = false;
    }
}

void minimize_instances ();

void add_isa_list (obj, classes, newp)
    KBObject* obj;
    KBList* classes;
    Boolean newp;
{
    for (; classes; classes = classes->next)
    {
	add_isa_parent (obj, classes->object, newp);
	newp = false;
    }
    minimize_instances (obj);
}


void clear_isa_parents (obj)
    KBObject* obj;
{
    KBList* isa;
    for (isa = obj->isa; isa; isa = isa->next)
    {
	remove (obj, isa->object->children);
    }
    free_list (obj->isa);
    obj->isa = nil;
}


void clear_instance_of (obj)
    KBObject* obj;
{
    KBList* inst_of;
    for (inst_of = obj->inst_of; inst_of; inst_of = inst_of->next)
    {
	remove (obj, inst_of->object->instances);
    }
    free_list (obj->inst_of);
    obj->inst_of = nil;
}

/* make sure object's isa list is minimal */

void minimize_object_isa (obj)
     KBObject* obj;
{
  KBList* old_isa_list;
  KBList* isa_list;
  KBList* ptr;

  old_isa_list = obj->isa;
  isa_list = minimize_classes (old_isa_list, false);
  /* remove link from old parent */
  for (ptr = old_isa_list; ptr; ptr = ptr->next)
    {
      if (! list_find (ptr->object, isa_list))
	{
	  remove (obj, ptr->object->children);
	}
    }
  /* add link for new parents */
  for (ptr = isa_list; ptr; ptr = ptr->next)
    {
      if (! list_find (ptr->object, old_isa_list))
	{
	  push (obj, ptr->object->children);
	}
    }

  obj->isa = isa_list;
  free_list (old_isa_list);

}

/* recurse through subclasses minimizing the instance lists of every instance */

void minimize_instances1 (obj, seen)
     KBObject* obj;
     int seen;
{
  KBList* instances;
  KBList* inst_list;
  KBList* old_inst_list;
  KBList* ptr;
  KBObject* inst;

  if (obj->seen == seen)
    return;

  obj->seen = seen;

  for (instances = obj->instances; instances; instances = instances->next)
    {
      inst = instances->object;
      old_inst_list = inst->inst_of;
      inst_list = minimize_classes (old_inst_list, false);
      /* remove link from old parent */
      for (ptr = old_inst_list; ptr; ptr = ptr->next)
	{
	  if (! list_find (ptr->object, inst_list))
	    {
	      remove (inst, ptr->object->instances);
	    }
	}
      /* add link for new parents */
      for (ptr = inst_list; ptr; ptr = ptr->next)
	{
	  if (! list_find (ptr->object, old_inst_list))
	    {
	      push (inst, ptr->object->instances);
	    }
	}

      inst->inst_of = inst_list;
      free_list (old_inst_list);
    }

  for (ptr = obj->children; ptr; ptr = ptr->next)
    {
      if (ptr->object != obj)
	{
	  minimize_instances1 (ptr->object, seen);
	  minimize_object_isa (ptr->object);
	}
    }
}

void minimize_instances (obj)
     KBObject* obj;
{
  unique_integer++;
  minimize_instances1 (obj, unique_integer);
}




/* compute the level of an object */

long get_level (obj)
    KBObject* obj;
{
    long level;

    if (! obj)
	return L_UNKNOWN;

    level = LEVELOF (obj);
    if (! level)
    {
	if (is_instance_of (obj, token_class))
	    level = L_TOKEN;
	else if (is_instance_of (obj, sclass_class))
	    level = L_SCLASS;
	else if (is_instance_of (obj, m1class_class))
	    level = L_M1;
	else if (is_instance_of (obj, m2class_class))
	    level = L_M2;
	else if (is_instance_of (obj, m3class_class))
	    level = L_M3;
	else if (is_instance_of (obj, m4class_class))
	    level = L_M4;
	else if (is_instance_of (obj, attribute_class))
	    level = L_ATTRIBUTE;
	else if (is_instance_of (obj, class_class))
	    level = L_OMEGA;
	else
	    level = L_UNKNOWN;
    }

    return level;
}



/* find all most specialized categories */

KBList* find_cats_class (obj, depth, result)
    KBObject* obj;
    int depth;
    KBList* result;
{
    KBList* ptr;
    KBList* ptr2;
    Boolean hit;

    if (depth > 1000)
	return result;

    depth++;

    for (ptr = obj->attributes; ptr; ptr = ptr->next)
    {
	/* check for not there */
	hit = false;
	for (ptr2 = result; ! hit && ptr2; ptr2 = ptr2->next)
	{
	    if (subclass_of (ptr2->object, ptr->object, 0))
		hit = true;
	    else if (subclass_of (ptr->object, ptr2->object, 0))
	    {
		ptr2->object = ptr->object;
		hit = true;
	    }
	}
	if (! hit)
	    result = cons (ptr->object, result);
    }
    for (ptr = obj->isa; ptr; ptr = ptr->next)
    {
	result = find_cats_class (ptr->object, depth, result);
    }

    return result;
}

KBList* find_cats (obj, result)
    KBObject* obj;
    KBList* result;
{
    KBList* ptr;

    for (ptr = obj->inst_of; ptr; ptr = ptr->next)
    {
	result = find_cats_class (ptr->object, 0, result);
    }

    return result;
}



/* finds all most specialized matches */

KBList* resolve_cat_class (obj, name, depth, result, found)
    KBObject* obj;
    char* name;
    int depth;
    KBList* result;
    Boolean *found;
{
    KBList* ptr;
    KBList* ptr2;
    Boolean hit;

    if (depth > 1000)
	return result;

    depth++;

    for (ptr = obj->attributes; ptr; ptr = ptr->next)
    {
	if (labeleq (name, ptr->object->name))
	{
	    /* check for not there */
	    hit = false;
	    *found = true;
	    for (ptr2 = result; ! hit && ptr2; ptr2 = ptr2->next)
	    {
		if (subclass_of (ptr2->object, ptr->object, 0))
		    hit = true;
		else if (subclass_of (ptr->object, ptr2->object, 0))
		{
		    ptr2->object = ptr->object;
		    hit = true;
		}
	    }
	    if (! hit)
		result = cons (ptr->object, result);
	}
    }
    for (ptr = obj->isa; ptr; ptr = ptr->next)
    {
	result = resolve_cat_class (ptr->object, name, depth, result, found);
    }

    return result;
}


KBList* resolve_cat (obj, name, result)
    KBObject* obj;
    char* name;
    KBList* result;
{
    KBList* ptr;
    Boolean found;

    found = false;
    for (ptr = obj->inst_of; ptr; ptr = ptr->next)
    {
	result = resolve_cat_class (ptr->object, name, 0, result, &found);
    }
    if (! found && obj->inst_of)
    {
	/* create an appropriate category in the first parent*/
	KBList metacats;

	metacats.object = attribute_class;
	metacats.next = nil;
	create_attr (obj->inst_of->object, name, &metacats, proposition_class);
	result = resolve_cat_class (obj->inst_of->object, name, 0, result, &found);
    }

    return result;
}

/* see if the attribute is instance of categories having the same name as objects */

Boolean attr_inst_of_isa_parent (class, obj_name, depth)
    KBObject* class;
    char* obj_name;
    int depth;
{
    KBList* ptr;
    int new_depth;

    if (depth > 1000)
	return false;

    new_depth = depth + 1;

    if (streq (class->name, obj_name))
	return true;

    for (ptr = class->isa; ptr; ptr = ptr->next)
    {
	if (attr_inst_of_isa_parent (ptr->object, obj_name, new_depth))
	    return true;
    }
    return false;
}


Boolean attr_inst_of (attr, obj_name)
    KBObject* attr;
    char* obj_name;
{
    KBList* parents;

    for (parents = attr->inst_of; parents; parents = parents->next)
    {
	if (! attr_inst_of_isa_parent (parents->object, obj_name, 0))
	    return false;
    }
    return true;
}

Boolean attr_inst_of_list (attr, objects)
    KBObject* attr;
    KBList* objects;
{
    KBList* ptr;

    for (ptr = objects; ptr; ptr = ptr->next)
    {
	if (! (attr_inst_of (attr, ptr->object->name)))
	    return false;
    }

    return true;
}




void fix_attr_isa (obj, attr, newp)
    KBObject* obj;
    KBObject* attr;
    Boolean   newp;
{
    KBList* parent_attrs;
    KBList* ptr;
    Boolean found;

    if (! attr->name)
	return;

    parent_attrs = nil;

    for (ptr = obj->isa; ptr; ptr = ptr->next)
    {
	parent_attrs = resolve_cat_class (ptr->object, attr->name, 0, parent_attrs, &found);
    }
    add_isa_list (attr, parent_attrs, newp);
    free_list (parent_attrs);
}

/* add all domains of cats to instances parents of val */

void verify_attr_val_types (val, cats, newobj)
    KBObject* val;
    KBList* cats;
    Boolean newobj;
{
    for (; cats; cats = cats->next)
    {
	if (cats->object->value && ! is_instance_of (val, cats->object->value))
	{
	    add_instance_of (val, cats->object->value, newobj);
	}
    }
}


void create_attr (obj, label, cats, val)
    KBObject* obj;
    char* label;
    KBList* cats;
    KBObject* val;
{
    KBObject* attr;
    KBList* ptr;
    Boolean new_attr;

    attr = nil;
    new_attr = false;
    if (label && *label != 0)
    {
	for (ptr = obj->attributes; ptr; ptr = ptr->next)
	{
	    if (ptr->object->name && streq (label, ptr->object->name))
	    {
		attr = ptr->object;
		break;
	    }
	}
    }
    if (! attr)
    {
	attr = new_object (label, obj);
	new_attr = true;
    }

    clear_isa_parents (attr);
    clear_instance_of (attr);
    attr->value = val;
    if (new_attr)
	pushf (attr, val->back_ptr);
    else
	push (attr, val->back_ptr);
    add_instance_of_list (attr, cats, new_attr);
    if (val->status & UNDEFINED)
	verify_attr_val_types (val, cats, false);
    if (new_attr)
	pushf (attr, obj->attributes);
    else
	push (attr, obj->attributes);
    fix_attr_isa (obj, attr, new_attr);
}


/* look up a name and complain to interpreter if it's not there and there is no */
/* connection to a knowledge base server*/

KBObject* lookup_name (clientData, interp, name, routine)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
    char* routine;
{
    KBObject* obj;

    if (! routine)
	routine = "";

    obj = get_object (name);

    if (! obj && tds_connection_exists)
    {
	Tcl_ResetResult (interp);
	Tcl_AppendResult (interp, routine, ": No such object '", name, "'", 0);
    }
    return obj;
}

/* look up and comfirm a name and complain to interpreter if it's not there */

KBObject* lookup_required_name (clientData, interp, name, routine)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
    char* routine;
{
    KBObject* obj;

    if (! routine)
	routine = "";

    CONFIRM (clientData, interp, name, routine);
    obj = get_object (name);
    if (! obj)
    {
	Tcl_ResetResult (interp);
	Tcl_AppendResult (interp, routine, ": No such object '", name, "'", 0);
    }
    return obj;
}



/* the command is */
/* kb attr obj cat1 , cat2 , cat3 label : val1 val2 val3 ... */

/* have to check for improving push's with knowledge of whether object is */
/* new or not */

int  process_attr_def (interp, obj, argc, argv)
    Tcl_Interp* interp;
    KBObject* obj;
    int argc;
    char* argv[];
{
    KBList* cats;
    KBList* old_cats;
    KBList* ptr;
    KBObject* val;
    int place;
    Boolean colon;
    char* label;
    KBObject* attr;

    label = nil;
    place = 3;
    colon = false;
    cats = nil;
    old_cats = nil;
    while (! colon)
    {
	/* looking at a cat */
	cats = resolve_cat (obj, argv[place], cats);
	if (cats == old_cats)
	{
	    Tcl_AppendResult (interp, "No such category '", argv[place], "' for ",
			      obj->name, (char*) 0);
	    return TCL_ERROR;
	}
	old_cats = cats;
	if (place + 1 >= argc)
	{
	    Tcl_AppendResult (interp, "No values attribute of ", obj->name, (char*) 0);
	    return TCL_ERROR;
	}
	if ((colon = streq (argv[place + 1], ":")) || streq (argv[place + 1], ","))
	{
	}
	else
	{
	    label = argv[place + 1];
	    place++;
	    colon = true;
	}
	place += 2;
	if (place >= argc)
	{
	    Tcl_AppendResult (interp, "No values attribute of ", obj->name, (char*) 0);
	    return TCL_ERROR;
	}
    }
    while (place < argc)
    {
	attr = new_object (label, obj);
	val = new_object (argv[place], nil);
	attr->value = val;
	push (attr, val->back_ptr);
	add_instance_of_list (attr, cats, false);
	push (attr, obj->attributes);
	verify_attr_val_types (val, cats, false);
	fix_attr_isa (obj, attr, false);
	place++;
    }
    free_list (cats);
    return TCL_OK;
}



/* the command is */
/* kb attr2 obj cats label val */

int  process_attr_def2 (interp, obj, argc, argv)
    Tcl_Interp* interp;
    KBObject* obj;
    int argc;
    char* argv[];
{
    KBList* cats;
    KBList* old_cats;
    KBList* ptr;
    KBObject* val;
    int place;
    Boolean colon;
    char* label;
    KBObject* attr;
    char**  cat_names;
    int  cat_count;
    int i;
    char** val_names;
    int  val_count;

    if (argc < 6)
    {
	interp->result = "Too few args for kb attr2";
	return TCL_ERROR;
    }

    label = argv[4];
    if (streq (label, "") || streq (label, "{}"))
	label = nil;
    colon = false;
    cats = nil;
    old_cats = nil;

    if (TCL_OK == Tcl_SplitList (interp, argv[3], &cat_count, &cat_names))
    {
	for (place = 0; place < cat_count; place++)
	{
	    cats = resolve_cat (obj, cat_names[place], cats);
	    if (cats == old_cats)
	    {
		/* Tcl_AppendResult (interp, "No such category '", cat_names[place], "' for ",
				  obj->name, (char*) 0);
		return TCL_ERROR; */
	    }
	    old_cats = cats;
	}
/* WINGS BEGIN */
/*	free (cat_names); */
/* WINGS END */
    }
    else
	return TCL_ERROR;


    for (place = 5; place < argc; place++)
    {
	if (*argv[place] == '"')
	{
	    attr = new_object (label, obj);
	    val = new_object (argv[place], nil);
	    attr->value = val;
	    push (attr, val->back_ptr);
	    add_instance_of_list (attr, cats, false);
	    verify_attr_val_types (val, cats, false);
	    push (attr, obj->attributes);
	    fix_attr_isa (obj, attr, false);
	}
	else if (TCL_OK == Tcl_SplitList (interp, argv[place], &val_count, &val_names))
	{
	    for (i = 0; i < val_count; i++)
	    {
		attr = new_object (label, obj);
		val = new_object (val_names[i], nil);
		attr->value = val;
		push (attr, val->back_ptr);
		add_instance_of_list (attr, cats, false);
		verify_attr_val_types (val, cats, false);
		push (attr, obj->attributes);
		fix_attr_isa (obj, attr, false);
	    }
	    place++;
/* WINGS BEGIN */
/*	    free (val_names); */
/* WINGS END */
	}
    }
    free_list (cats);
    return TCL_OK;
}


void interp_result_list_names (interp, list)
    Tcl_Interp* interp;
    KBList* list;
{
    for (; list; list = list->next)
	Tcl_AppendElement (interp, list->object->name);
}


void interp_result_list_ids (interp, list)
    Tcl_Interp* interp;
    KBList* list;
{
    for (; list; list = list->next)
	Tcl_AppendElement (interp, list->object->id);
}

/* WINGS BEGIN */
/* kb destroy obj */

int kb_destroy (clientData, interp, name)
	ClientData clientData;
	Tcl_Interp* interp;
	char* name;
{
	KBObject* obj;

	obj = lookup_name (clientData, interp, name, "kb destroy");
	if (! obj)
		return TCL_ERROR;

	Tcl_DeleteHashEntry(Tcl_FindHashEntry(&kb_symbols, name));

	return TCL_OK;
}

/* kb exists obj */

int kb_exists (clientData, interp, name)
	ClientData clientData;
	Tcl_Interp* interp;
	char* name;
{
    KBObject* obj;

    if (tds_connection_exists)
    {
	if (! tmb_exists (clientData, interp, name))
	{
	    obj = get_object (name);
	}
	else
	{
	    CONFIRM (clientData, interp, name, "kb exists");
	    obj = get_object (name);
	}
    }
    else
    {
	CONFIRM (clientData, interp, name, "kb exists");
	obj = get_object (name);
    }
    if (obj)
    {
	interp->result = "1";
	return 1;
    }
    else
    {
	interp->result = "0";
	return 0;
    }
}
/* WINGS END */

/* kb level obj */

int kb_do_level (clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp* interp;
    int argc;
    char* argv[];
{
    KBObject* obj;

    if (argc != 3)
    {
	interp->result = "Wrong number of args for test: kb test [isa|inst*] obj1 obj2";
	return TCL_ERROR;
    }

    obj = lookup_name (clientData, interp, argv[2], "kb level");
    if (! obj)
	return TCL_ERROR;

    switch (get_level (obj))
    {
      case L_OMEGA:
	interp->result = "OmegaClass";
	break;
      case L_TOKEN:
	interp->result = "Token";
	break;
      case L_SCLASS:
	interp->result = "SimpleClass";
	break;
      case L_M1:
	interp->result = "MetaClass";
	break;
      case L_M2:
	interp->result = "MetaMetaClass";
	break;
      case L_M3:
	interp->result = "MetaMetaMetaClass";
	break;
      case L_M4:
	interp->result = "MetaMetaMetaMetaClass";
	break;
      case L_ATTRIBUTE:
	interp->result = "AttributeClass";
	break;
      default:
	interp->result = "Unknown";
	break;
    }

    return TCL_OK;
}

Boolean only_close_quote (str)
    char* str;
{
    for (; *str; str++)
    {
	if (*str == '"' && *(str + 1) != 0)
	    return false;
    }
    return true;
}


static char* special_chars = "\";: \t\r\n[]{}!@#$-+=`~,./\\";

/* print object name or id to str, escaping if it has special characters */
/* if a string is okay in position, don't escape, otherwise escape	 */

void	kb_escape_obj (obj, str, notstring)
    KBObject* obj;
    FILE*     str;
    Boolean   notstring;
{
    Boolean special;
    char* ptr;
    char* sc_ptr;

    if (! obj)
	fprintf (str, "NULL");
    else if (! obj->name)
	fprintf (str, "|%s|", obj->id);
    else
    {
	if (! notstring
	    && *(obj->name) == '"'
	    && only_close_quote (obj->name + 1))
	{
	    fprintf (str, "%s", obj->name);
	}
	else
	{
	    special = false;
	    for (ptr = obj->name; ! special && *ptr; ptr++)
	    {
		for (sc_ptr = special_chars; ! special && *sc_ptr; sc_ptr++)
		{
		    if (*ptr == *sc_ptr)
		    {
			special = true;
		    }
		}
	    }
	    if (special)
		fprintf (str, "|%s|", obj->name);
	    else
		fprintf (str, "%s", obj->name);
	}
    }
}

/* dump an object to a file */

int	kb_dump_object (clientData, interp, obj, str)
    ClientData	clientData;
    Tcl_Interp* interp;
    KBObject*	obj;
    FILE*	str;
{
    KBList* ptr;
    KBList* in_ptr;
    KBObject* val;
    KBObject* attr;
    KBObject* cat;
    KBList* last_cat;
    char* comma;

    if (is_instance_of (obj, m4class_class))
	fprintf (str, "\nM4Class");
    else if (is_instance_of (obj, m3class_class))
	fprintf (str, "\nM3Class");
    else if (is_instance_of (obj, m2class_class))
	fprintf (str, "\nMetaMetaClass");
    else if (is_instance_of (obj, m1class_class))
	fprintf (str, "\nMetaClass");
    else if (is_instance_of (obj, sclass_class))
	fprintf (str, "\nSimpleClass");
    else if (is_instance_of (obj, token_class))
	fprintf (str, "\nToken");
    else
	fprintf (str, "\nClass");

    fprintf (str, " ");
    kb_escape_obj (obj, str, 1);
    fprintf (str, "\n    in");
    if (obj->inst_of)
    {
	comma = "";
	for (ptr = obj->inst_of; ptr; ptr = ptr->next)
	{
	    val = ptr->object;
	    fprintf (str, "%s ", comma);
	    kb_escape_obj (val, str, 1);
	    comma = ",";
	}
    }
    else
    {
	fprintf (str, " Class");
    }
    if (obj->isa)
    {
	fprintf (str, "\n    isa");
	comma = "";
	for (ptr = obj->isa; ptr; ptr = ptr->next)
	{
	    val = ptr->object;
	    fprintf (str, "%s ", comma);
	    kb_escape_obj (val, str, 1);
	    comma = ",";
	}
    }
    if (obj->attributes)
    {

	fprintf (str, "\n    with");
	last_cat = nil;
	for (ptr = obj->attributes; ptr; ptr = ptr->next)
	{
	    attr = ptr->object;
	    val = attr->value;
	    /* WINGS BEGIN */
	    if (kb_exists(clientData, interp, val->name))
	    {
	    if (list_equal (last_cat, attr->inst_of))
	    {
		fprintf (str, ";");
	    }
	    else
	    {
		fprintf (str, "\n      ");
		comma = "";
		for (in_ptr = attr->inst_of; in_ptr; in_ptr = in_ptr->next)
		{
		    cat = in_ptr->object;
		    fprintf (str, "%s", comma);
		    kb_escape_obj (cat, str, 1);
		    comma = ", ";
		}
		last_cat = attr->inst_of;
	    }
	    if (attr->name)
	    {
		fprintf (str, "\n\t");
		kb_escape_obj (attr, str, 1);
		fprintf (str, " :");
	    }
	    else
		fprintf (str, "\n\t:");
	    fprintf (str, " ");
	    kb_escape_obj (val, str, 0);
	    }
	    /* WINGS END */
	}
    }
    fprintf (str, "\nend\n");
}



/* dump the knowledge base to a file */

int	kb_dump_file (clientData, interp, file_name)
    ClientData clientData;
    Tcl_Interp* interp;
    char*   file_name;
{
    FILE  *outfile;
    int i;
    KBObject* obj;

    if (! file_name)
    {
	interp->result = "Empty file name to dump";
	return TCL_ERROR;
    }
    outfile = fopen (file_name, "w");
    if (! outfile)
    {
	perror ("kb dump open");
	Tcl_AppendResult (interp, "kb dump couldn't open file ", file_name, 0);
	return TCL_ERROR;
    }
    for (i = dump_start; i < obj_id; i++)
    {
	obj = *(all_objects + i);
	if (! obj->attribute_of
	    && ! is_instance_of (obj, number_class)
	    && ! is_instance_of (obj, string_class)
	    && ! is_instance_of (obj, pstring_class)
	    && kb_exists (clientData, interp, obj->name))
	{
	    kb_dump_object (clientData, interp, obj, outfile);
	}
    }

    fclose (outfile);

    return TCL_OK;
}




/* kb test [isa|inst*] [any|all] obj1 class1 class2 ... */

int kb_do_test (clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp* interp;
    int argc;
    char* argv[];
{
    KBObject* obj1;
    KBObject* class;
    int argn;
    int i;
    Boolean any;

    if (argc < 5)
    {
	interp->result = "Wrong number of args for test: kb test [isa|inst*] [any|all] obj1 obj2";
	return TCL_ERROR;
    }
    if (strcaseeq (argv[3], "any"))
    {
	argn = 4;
	any = true;
    }
    else if (strcaseeq (argv[3], "all"))
    {
	argn = 4;
	any = false;
    }
    else
    {
	argn = 3;
	any = false;
    }
    obj1 = lookup_required_name (clientData, interp, argv[argn], "kb test");
    if (! obj1)
	return TCL_ERROR;

    if (streq (argv[2], "isa"))
    {
	if (any)
	{
	    interp->result = "0";
	    for (i = argn+1; i < argc; i++)
	    {
		class = lookup_required_name (clientData, interp, argv[i], "kb test");
		if (! class)
		    return TCL_ERROR;

		if (subclass_of (obj1, class, 0))
		{
		    interp->result = class->id;
		    break;
		}
	    }
	}
	else
	{
	    interp->result = "1";
	    for (i = argn+1; i < argc; i++)
	    {
		class = lookup_required_name (clientData, interp, argv[i], "kb test");
		if (! class)
		    return TCL_ERROR;

		if (! subclass_of (obj1, class, 0))
		{
		    interp->result = "0";
		    break;
		}
	    }
	}
    }
    else if (strncmp (argv[2], "inst", 4) == 0)
    {
	if (any)
	{
	    interp->result = "0";
	    if (obj1->attribute_of)
	    {
		KBList* classes = nil;
		KBList* ptr;
		for (i = argn+1; i < argc; i++)
		{
		    classes = resolve_cat (obj1->attribute_of, argv[i], classes);
		    for (ptr = classes; ptr; ptr = ptr->next)
		    {
			if (is_instance_of (obj1, ptr->object))
			{
			    interp->result = ptr->object->id;
			    free_list (classes);
			    return TCL_OK;
			}
		    }
		}
		free_list (classes);
	    }
	    else
	    {
		for (i = argn+1; i < argc; i++)
		{
		    class = lookup_required_name (clientData, interp, argv[i], "kb test");
		    if (! class)
			return TCL_ERROR;

		    if (is_instance_of (obj1, class))
		    {
			interp->result = class->id;
			break;
		    }
		}
	    }
	}
	else
	{
	    interp->result = "1";
	    if (obj1->attribute_of)
	    {
		KBList* classes = nil;
		KBList* ptr;
		for (i = argn+1; i < argc; i++)
		{
		    classes = resolve_cat (obj1->attribute_of, argv[i], classes);
		    for (ptr = classes; ptr; ptr = ptr->next)
		    {
			if (! is_instance_of (obj1, ptr->object))
			{
			    interp->result = "0";
			    free_list (classes);
			    return TCL_OK;
			}
		    }
		}
		free_list (classes);
	    }
	    else
	    {
		for (i = argn+1; i < argc; i++)
		{
		    class = lookup_required_name (clientData, interp, argv[i], "kb test");
		    if (! class)
			return TCL_ERROR;

		    if (! is_instance_of (obj1, class))
		    {
			interp->result = "0";
			break;
		    }
		}
	    }
	}
    }
    else
    {
	Tcl_AppendResult (interp, "kb test: bad test ", argv[2], 0);
	return TCL_ERROR;
    }

    return TCL_OK;
}


int recursive_instances1 (clientData, interp, obj, seen)
    ClientData clientData;
    Tcl_Interp* interp;
    KBObject* obj;
    int seen;
{
    KBList* ptr;

    if (seen == obj->seen)
      return TCL_OK;

    obj->seen = unique_integer;

    for (ptr = obj->instances; ptr; ptr = ptr->next)
    {
	Tcl_AppendElement (interp, ptr->object->id);
    }
    CONFIRM_CHILDREN_O (clientData, interp, obj);
    for (ptr = obj->children; ptr; ptr = ptr->next)
    {
	CONFIRM_INSTANCES_O (clientData, interp, ptr->object);
	recursive_instances1 (clientData, interp, ptr->object, seen);
    }

    return TCL_OK;
}


int recursive_instances (clientData, interp, name)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
{
    KBObject* obj;

    CONFIRM_INSTANCES (clientData, interp, name);
    obj = lookup_name (clientData, interp, name, "kb instances -rec");
    if (! obj)
	return TCL_ERROR;

    unique_integer++;

    return recursive_instances1 (clientData, interp, obj, unique_integer);
}


static Tcl_DString compose_result_string;

Tcl_DString* compose_rep_args (argc, argv)
    int argc;
    char** argv;
{
    int i;
    KBObject* obj;

    Tcl_DStringFree (&compose_result_string);
    for (i = 0; i < argc; i++)
    {
	obj = get_object (argv[i]);
	if (i > 0)
	    Tcl_DStringAppend (&compose_result_string, " ", 1);
	if (obj && obj->name)
	    Tcl_DStringAppend (&compose_result_string, obj->name, -1);
	else
	    Tcl_DStringAppend (&compose_result_string, argv[i], -1);
    }


    return &compose_result_string;
}

KBObject* ckb_new_object (name, instance_of, newp)
    char* name;
    char* instance_of;
    int*  newp;
{
    KBObject* obj;

    obj = new_object (name, nil);
    *newp = (obj->status & UNDEFINED);
    obj->status &= ~UNDEFINED;
    if (instance_of)
	add_instance_of (obj, new_object (instance_of, nil), *newp);

    return obj;
}

int ckb_defined (name)
{
    KBObject* obj = get_object (name);
    if (!obj || (obj->status & UNDEFINED))
	return 0;
    else
	return 1;
}


int kb_match_names (interp, pattern)
    Tcl_Interp* interp;
    char* pattern;
{
    int i;
    int status;
    KBObject* obj;
    Boolean strings = pattern ? (*pattern == '"') : false;
    Tcl_RegExp exp = pattern ? Tcl_RegExpCompile (interp, pattern) : 0;

    if (! exp)
	return TCL_ERROR;

    if (strings)
    {
	for (i = dump_start; i < obj_id; i++)
	{
	    obj = *(all_objects + i);
	    if (obj->name && (status = Tcl_RegExpExec (interp, exp, obj->name, obj->name)) == 1)
	    {
		Tcl_AppendElement (interp, obj->name);
	    }
	    else if (status == -1)
		return TCL_ERROR;
	}
    }
    else
    {
	for (i = dump_start; i < obj_id; i++)
	{
	    obj = *(all_objects + i);
	    if (obj->name && *(obj->name) != '"'
		&& (status = Tcl_RegExpExec (interp, exp, obj->name, obj->name)) == 1)
	    {
		Tcl_AppendElement (interp, obj->name);
	    }
	    else if (status == -1)
		return TCL_ERROR;
	}
    }

    return TCL_OK;
}



# define    ARGCHECK(min,max,command) { if (argc < (min) || ((max) > 0 && argc > (max))) { Tcl_AppendResult (interp, "Wrong number of args; kb ", (command), 0); return TCL_ERROR; }}

extern void             RepInProc ();



int tmb_exists (clientData, interp, name)
    ClientData clientData;
    Tcl_Interp* interp;
    char* name;
{
     if (tds_connection_exists)
     {
	  if (! name)
	  {
	       interp->result = "No name";
	       return TCL_ERROR;
	  }
	  RepGeneric (clientData, interp, "exists", name, nil);
	  return (TCL_OK == rep_error_code);
     }
     else
	  return 0;
}



void init_kb ();

int parse_rule ();
int do_goals ();
int do_solve ();
extern int chain_trace_level;

/* form one is kb key obj [val1 val2 ... ] */
void* get_tm_command ();

int kb_comm (clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp* interp;
    int argc;
    char* argv[];
{
    int i;
    int j;
    KBObject* obj;
    KBObject* other_obj;
    KBObject* cat;
    KBObject* val;
    KBList* ptr;
    KBList* new_ptr;
    KBList* cats;
    char**  class_names;
    int	    class_count;
    int arg_len;
    char*   label;
    Boolean found;

    if (argc >= 2)
	arg_len = strlen (argv[1]);
    else
	arg_len = 0;

    if (argc < 3 && (streq (argv[1], "size") || ! get_tm_command (argv[1])))
    {
	if (argc == 2 && streq (argv[1], "init"))
	    init_kb ();
	else if (argc == 2 && streq (argv[1], "size"))
	{
	    sprintf (interp->result, "%d", obj_id);
	}
	else
	{
	    interp->result = "Wrong # args";
	    return TCL_ERROR;
	}
    }
    else if (streq (argv[1], "nth"))
    {
	if (argc != 3)
	{
	    interp->result = "Wrong # args to kb nth";
	    return TCL_ERROR;
	}
	if (TCL_OK != Tcl_GetInt (interp, argv[2], &i))
	    return TCL_ERROR;
	obj = ith_object (i);
	if (obj)
	    interp->result = obj->name ? obj->name : obj->id;
	else
	{
	    interp->result = "No nth object";
	    return TCL_ERROR;
	}
    }
    else if (streq (argv[1], "create"))
    {
	char* name = argv[2] == "" ? 0 : argv[2];
	obj = new_object (name, nil);
	clear_isa_parents (obj);
	clear_instance_of (obj);
	interp->result = obj->id;
	for (i = 3; i < argc; i++)
	{
	    if (TCL_OK == Tcl_SplitList (interp, argv[i], &class_count, &class_names))
	    {
		for (j = 0; j < class_count; j++)
		{
		    other_obj = new_object (class_names[j], nil); /* always returns something */
		    add_instance_of (obj, other_obj, false);
		}
/* WINGS BEGIN */
/*		free (class_names); */
/* WINGS END */
	    }
	}
    }
/* WINGS BEGIN */
    else if (streq (argv[1], "destroy"))
    {
	if (argc != 3)
	{
		interp->result = "Wrong number of args for destroy: kb destroy obj";
		return TCL_ERROR;
	}
    	return kb_destroy (clientData, interp, argv[2]);
    }
/* WINGS END */
    else if (streq (argv[1], "level"))
    {
	return kb_do_level (clientData, interp, argc, argv);
    }
    else if (streq (argv[1], "test"))
    {
	return kb_do_test (clientData, interp, argc, argv);
    }
    else if (arg_len >= 3 && strncmp (argv[1], "current", arg_len) == 0)
    {
	if (strncmp (argv[2], "source", strlen (argv[2])) == 0)
	{
	    if (argc == 3)
	    {
	    }
	    else if (argc == 4 && strncmp (argv[3], "default", strlen (argv[3])) == 0)
	    {
		current_source = default_source;
	    }
	    else if (argc == 5)
	    {
		if (strncmp (argv[3], "file", strlen (argv[3])) == 0)
		{
		    /* should really look for existing file source */
		    current_source = new (Source);
		    current_source->source_type = SOURCE_FILE;
		    current_source->source.file.name = strdup (argv[4]);
		}
		else
		{
		    interp->result = "Error: can only set file (kb current source [default | type value])";
		    return TCL_ERROR;
		}
	    }
	    else
	    {
		interp->result = "Error: kb current source [default | type value]";
		return TCL_ERROR;
	    }
	    Tcl_AppendElement (interp, source_names[current_source->source_type]);
	    switch (current_source->source_type)
	    {
	      case SOURCE_FILE:
		Tcl_AppendElement (interp, current_source->source.file.name);
		break;

	      default:
		Tcl_AppendElement (interp, "");
	    }
	}
	else
	{
	    interp->result = "Error: kb current source [default | type value]";
	    return TCL_ERROR;
	}
    }
    else if (arg_len >= 3 && strncmp (argv[1], "source", arg_len) == 0)
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb source");

	if (! obj)
	    return TCL_ERROR;

	if (argc == 4)
	{
	    interp->result = "Modify source not implemented";
	    return TCL_ERROR;
	}
	else if (argc > 4)
	{
	    interp->result = "Too many arguments: kb source obj";
	    return TCL_ERROR;
	}
	Tcl_AppendElement (interp, source_names[obj->source->source_type]);
	switch (obj->source->source_type)
	{
	  case SOURCE_FILE:
	    Tcl_AppendElement (interp, obj->source->source.file.name);
	    break;

	  default:
	    Tcl_AppendElement (interp, "");
	}
    }
    else if (strncmp (argv[1], "mod", 3) == 0)
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb modified");

	if (! obj)
	    return TCL_ERROR;

	if (argc == 4)
	{
	    /* set modified flag */
	    Boolean mod;
	    if (Tcl_GetBoolean (interp, argv[3], &mod) != TCL_OK)
		return TCL_ERROR;
	    if (mod)
		obj->status |= USER_MODIFIED;
	    else
		obj->status &= ~USER_MODIFIED;
	}
	else if (argc > 4)
	{
	    interp->result = "Too many arguments: kb modified obj [boolean]";
	    return TCL_ERROR;
	}
	if (STATUS (obj, USER_MODIFIED))
	    interp->result = "1";
	else
	    interp->result = "0";
    }
    else if (streq (argv[1], "exists"))
    {
	kb_exists (clientData, interp, argv[2]);
    }
    else if (strcaseeq (argv[1], "rule"))
	return parse_rule (clientData, interp, argc, argv);
    else if (strncmp (argv[1], "goal", 4) == 0)
	return do_goals (clientData, interp, argc, argv);
    else if (strncmp (argv[1], "solve", 4) == 0)
	return do_solve (clientData, interp, argc, argv);
    else if (streq (argv[1], "instance_names"))
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb instance_names");

	if (! obj)
	    return TCL_ERROR;

	if (argc == 3)
	{
	    /* wants us to return list of instance_of */
	    if (! obj->inst_of)
		Tcl_AppendElement (interp, PROP);
	    else
		interp_result_list_names (interp, obj->inst_of);
	}
	else
	{
	    interp->result = "Wrong number of args: kb instance_names obj";
	    return TCL_ERROR;
	}
    }
    else if (streq (argv[1], "trace"))
    {
	if (TCL_OK != Tcl_GetInt (interp, argv[2], &i))
	{
	    sprintf (interp->result, "kb trace: bad number %s", argv[2]);
	    return TCL_ERROR;
	}
	chain_trace_level = i;
    }
    else if (streq (argv[1], "debug_tds"))
    {
	if (TCL_OK != Tcl_GetInt (interp, argv[2], &i))
	{
	    sprintf (interp->result, "kb debug_tds: bad number %s", argv[2]);
	    return TCL_ERROR;
	}
	debug_conn = i;
    }
    else if (streq (argv[1], "instance_of"))
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb instance_of");
	if (! obj)
	    return TCL_ERROR;

	if (argc < 4)
	{
	    /* wants us to return list of instance_of */
	    if (! obj->inst_of)
		Tcl_AppendElement (interp, PROP);
	    else
		interp_result_list_ids (interp, obj->inst_of);
	}
	else
	{
	    for (i = 3; i < argc; i++)
	    {
		if (TCL_OK == Tcl_SplitList (interp, argv[i], &class_count, &class_names))
		{
		    for (j = 0; j < class_count; j++)
		    {
			other_obj = new_object (class_names[j], nil); /* always returns something */
			add_instance_of (obj, other_obj, false);
		    }
/* WINGS BEGIN */
/*		    free (class_names); */
/* WINGS END */
		}
	    }
	}
    }
    else if (streq (argv[1], "instances"))
    {
	if (strncasecmp (argv[2], "-rec", 4) == 0)
	{
	    /* recursively collect instances for each arg */
	    for (i = 3; i < argc; i++)
	    {
		if (TCL_OK != recursive_instances (clientData, interp, argv[i]))
		    return TCL_ERROR;
	    }
	}
	else
	{
	    CONFIRM_INSTANCES (clientData, interp, argv[2]);
	    obj = lookup_name (clientData, interp, argv[2], "kb instances");
	    if (! obj)
	    {
		Tcl_AppendResult (interp, "No such object: '", argv[2], "'", 0);
		return TCL_ERROR;
	    }

	    if (argc < 4)
	    {
		/* wants us to return list of instances */
		interp_result_list_ids (interp, obj->instances);
	    }
	    else
	    {
		for (i = 3; i < argc; i++)
		{
		    other_obj = new_object (argv[i], nil); /* always returns something */
		    add_instance_of (other_obj, obj, false);
		}
	    }
	}
    }
    else if (streq (argv[1], "isa"))
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb isa");
	if (! obj)
	    return TCL_ERROR;

	if (argc < 4)
	{
	    /* wants us to return list of isa */
	    interp_result_list_ids (interp, obj->isa);
	}
	else
	{
	    for (i = 3; i < argc; i++)
	    {
		if (TCL_OK == Tcl_SplitList (interp, argv[i], &class_count, &class_names))
		{
		    for (j = 0; j < class_count; j++)
		    {
			other_obj = new_object (class_names[j], nil); /* always returns something */
			add_isa_parent (obj, other_obj, false);
		    }
/* WINGS BEGIN */
/*		    free (class_names); */
/* WINGS END */
		}
		minimize_instances (obj);
	    }
	}
    }
    else if (streq (argv[1], "children"))
    {
	obj = lookup_name (clientData, interp, argv[2], "kb children");
	CONFIRM_CHILDREN (clientData, interp, argv[2]);
	if (! obj)
	    return TCL_ERROR;

	if (argc < 4)
	{
	    /* wants us to return list of children */
	    interp_result_list_ids (interp, obj->children);
	}
	else
	{
	    for (i = 3; i < argc; i++)
	    {
		other_obj = new_object (argv[i], nil); /* always returns something */
		add_isa_parent (other_obj, obj, false);
		minimize_instances (other_obj);
	    }
	}
    }
    else if (streq (argv[1], "all_cats"))
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb all_cats");
	if (obj)
	{
	    ptr = find_cats (obj, nil);
	    interp_result_list_names (interp, ptr);
	    free_list (ptr);
	}
    }
    else if (strcaseeq (argv[1], "value_id")
	     || strcaseeq (argv[1], "val_id")
	     || strcaseeq (argv[1], "to"))
    {
	obj = get_object (argv[2]);
	if (obj && obj->value)
	    Tcl_AppendElement (interp, obj->value->id);
    }
    else if (strcaseeq (argv[1], "value")
	     || strcaseeq (argv[1], "val"))
    {
	obj = get_object (argv[2]);
	if (argc == 3)
	{
	    if (obj && obj->value)
	    {
		if (obj->value->name)
		    Tcl_AppendResult (interp, obj->value->name, 0);
		else
		    Tcl_AppendResult (interp, obj->value->id, 0);
	    }
	}
	else if (argc == 4)
	{
	    if (! obj)
	    {
		Tcl_AppendElement (interp, "No such object ");
		Tcl_AppendElement (interp, argv[2]);
		return TCL_ERROR;
	    }
	    val = new_object (argv[3], nil);
	    obj->value = val;
	}
	else
	{
	    Tcl_AppendResult (interp, "Wrong number of args; kb value obj [new value]", 0);
	    return TCL_ERROR;
	}
    }
    else if (strcaseeq (argv[1], "name")
	     || strcaseeq (argv[1], "label"))
    {
	CONFIRM (clientData, interp, argv[2], "kb name");
	if (argc == 3)
	{
	    obj = get_object (argv[2]);
	    if (obj)
	    {
		if (obj->name)
		    interp->result = obj->name;
	    }
	}
	else if (argc == 4)
	{
	    obj = get_object (argv[2]);
	    if (obj)
	    {
		rename_obj (obj, argv[3]);
	    }
	}
	else
	{
	    Tcl_AppendResult (interp, "Wrong number of args; kb name obj [new value]", 0);
	    return TCL_ERROR;
	}
    }
    else if (streq (argv[1], "full_name"))
    {
	obj = get_object (argv[2]);
	if (obj)
	{
	    if (obj->attribute_of)
	    {
		Tcl_AppendResult (interp,
				  (obj->attribute_of->name ? obj->attribute_of->name :
				   obj->attribute_of->id),
				  ".",
				  (obj->name ? obj->name : obj->id),
				  0);
	    }
	    else if (obj->name)
		interp->result = obj->name;
	    else
		interp->result = obj->id;
	}
    }
    else if (streq (argv[1], "request"))
    {
	Tcl_DString* arg_str;
	if (strncmp (argv[2], "-call", 4) == 0)
	{
	    ARGCHECK (5,0,"request [-callback callback] obj1 [obj2 ...]")
	    arg_str = compose_rep_args (argc - 4, argv + 4);
	    RepRequestObject (clientData, interp, Tcl_DStringValue (arg_str), argv[3]);
	}
	else
	{
	    ARGCHECK (3,0,"request [-callback callback] obj1 [obj2 ...]")
	    arg_str = compose_rep_args (argc - 2, argv + 2);
	    RepRequestObject (clientData, interp, Tcl_DStringValue (arg_str), nil);
	}
	Tcl_DStringResult (interp, &rep_result);
	if (rep_error_code != TCL_OK)
	    return rep_error_code;
    }
    else if (streq (argv[1], "ask"))
    {
	Tcl_DString* arg_str;
	if (strncmp (argv[2], "-call", 4) == 0)
	{
	    ARGCHECK (5,0,"ask [-callback callback] command [arg1 arg2 ...]")
	    arg_str = compose_rep_args (argc - 5, argv + 5);
	    RepAsk (clientData, interp, argv[4], Tcl_DStringValue (arg_str), argv[3]);
	}
	else
	{
	    ARGCHECK (3,0,"ask [-callback callback] command [arg1 arg2 ...]")
	    arg_str = compose_rep_args (argc - 3, argv + 3);
	    RepAsk (clientData, interp, argv[2], Tcl_DStringValue (arg_str), nil);
	}
	Tcl_DStringResult (interp, &rep_result);
	if (rep_error_code != TCL_OK)
	    return rep_error_code;
    }
    else if (streq (argv[1], "command"))
    {
	Tcl_DString* arg_str;
	if (strncmp (argv[2], "-call", 4) == 0)
	{
	    ARGCHECK (5,0,"command [-callback callback] command [arg1 arg2 ...]")
	    arg_str = compose_rep_args (argc - 5, argv + 5);
	    RepCommand (clientData, interp, argv[4], Tcl_DStringValue (arg_str), argv[3]);
	}
	else
	{
	    ARGCHECK (3,0,"command [-callback callback] command [arg1 arg2 ...]")
	    arg_str = compose_rep_args (argc - 3, argv + 3);
	    RepCommand (clientData, interp, argv[2], Tcl_DStringValue (arg_str), nil);
	}
	Tcl_DStringResult (interp, &rep_result);
	if (rep_error_code != TCL_OK)
	    return rep_error_code;
    }
/* WINGS BEGIN */
#ifdef WINGS_UNIX
    else if (streq (argv[1], "connect"))
    {
	char* host;
	int   port;
	int   fd;

	host = 0;
	port = 1111;
	if (argc > 2)
	    host = argv[2];
	if (argc > 3)
	    port = atoi (argv[3]);
	fd = tmb (host, port);
	if (fd)
	{
	    Tk_CreateFileHandler(fd, TK_READABLE, RepInProc, (ClientData) 0);
	}
	else
	{
	    Tcl_AppendResult (interp, "Couldn't connect to host", 0);
	    return TCL_ERROR;
	}
    }
#endif
/* WINGS END */
    else if (streq (argv[1], "load"))
    {
	for (i = 2; i < argc; i++)
	{
	    if (parse_file (argv[i]))
	    {
		sprintf (interp->result, "Parse error in %s", argv[i]);
		return TCL_ERROR;
	    }
	}
	interp->result = "1";
	return TCL_OK;
    }
    else if (streq (argv[1], "dump"))
    {
	if (argc == 3)
	    return kb_dump_file (clientData, interp, argv[2]);
	else
	{

	}
    }
/* WINGS BEGIN */
#ifdef WINGS_UNIX
    else if (streq (argv[1], "write"))
    {
	if (argc <= 3)
	{
	    Tcl_AppendResult (interp, "kb write: Too few arguments", 0);
	    return TCL_ERROR;
	}
	else
	{
	    int r1;
	    FILE*   str;
	    int j;
	    char** val_names;
	    int  val_count;

	    r1 = Tcl_GetOpenFile (interp, argv[2], 1, 1, &str);
	    if (r1 != TCL_OK)
		return r1;
	    for (i = 3; i < argc; i++)
	    {
		if (TCL_OK == Tcl_SplitList (interp, argv[i], &val_count, &val_names))
		{
		    for (j = 0; j < val_count; j++)
		    {
			obj = lookup_required_name (clientData, interp, val_names[j], "kb write");
			if (obj)
			    kb_dump_object (clientData, interp, obj, str);
			else
			    return TCL_ERROR;
		    }
		}
		else
		{
		    obj = lookup_required_name (clientData, interp, argv[i], "kb write");
		    if (obj)
			kb_dump_object (clientData, interp, obj, str);
		    else
			return TCL_ERROR;
		}
	    }
	}
    }
#endif
/* WINGS END */
    else if (streq (argv[1], "tell"))
    {
	for (i = 2; i < argc; i++)
	{
	    if (parse_string (argv[i]))
	    {
		sprintf (interp->result, "Parse error in %s", argv[i]);
		return TCL_ERROR;
	    }
	}
	interp->result = "1";
	return TCL_OK;
    }
    else if (streq (argv[1], "id"))
    {
	/* kb id obj1 obj2 ... */
	/* or */
	/* kb id label in obj  (returns all most specialized */
	if (argc == 5 && strcaseeq (argv[3], "in"))
	{
	    label = argv[2];
	    if (streq (label, "") || streq (label, "{}"))
		label = nil;
	    obj = get_object (argv[4]);
	    if (! obj)
	    {
		interp->result = "";
		return TCL_OK;
	    }
	    cats = resolve_cat_class (obj, label, 0, nil, &found);
	    for (ptr = cats; ptr; ptr = ptr->next)
	    {
		if (ptr->object)
		{
		    if (ptr->object->id)
			Tcl_AppendElement (interp, ptr->object->id);
		    else
			Tcl_AppendElement (interp, "");
		}
		else
		    Tcl_AppendElement (interp, "");
	    }
	    free_list (cats);
	}
	else
	{
	    for (i = 2; i < argc; i++)
	    {
		obj = get_object (argv[i]);
		if (obj)
		{
		    if (obj->id)
			Tcl_AppendElement (interp, obj->id);
		    else
			Tcl_AppendElement (interp, "");
		}
		else
		    Tcl_AppendElement (interp, "");
	    }
	}
    }
    else if (streq (argv[1], "back_ptr"))
    {
	obj = get_object (argv[2]);
	if (obj)
	    interp_result_list_ids (interp, obj->back_ptr);
    }
    else if (strcaseeq (argv[1], "attr_of")
	     || strcaseeq (argv[1], "from"))
    {
	obj = get_object (argv[2]);
	if (obj && obj->attribute_of)
	    interp->result = obj->attribute_of->id;
    }
    else if (strncmp (argv[1], "parse_a", 7) == 0)
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb parse_attr");
	if (! obj)
	    return TCL_ERROR;

	if (argc < 4)
	{
	    /* wants us to return list of attributes */
	    interp_result_list_ids (interp, obj->attributes);
	}
	else if (argc < 6)
	{
	    interp->result = "Wrong # args";
	    return TCL_ERROR;
	}
	else
	{
	    return process_attr_def (interp, obj, argc, argv);
	}
    }
    else if (streq (argv[1], "attr_names"))
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb attr_names");
	if (! obj)
	    return TCL_ERROR;

	if (argc < 4)
	{
	    /* wants us to return list of attributes */
	    interp_result_list_names (interp, obj->attributes);
	}
    }
    else if (streq (argv[1], "attr2")
	     ||streq (argv[1], "attributes")
	     || streq (argv[1], "attribute")
	     || streq (argv[1], "attr")
	     || streq (argv[1], "attrs"))
    {
	if (strncasecmp (argv[2], "-in", 3) == 0)
	{
	     obj = lookup_required_name (clientData, interp, argv[3], "kb attributes -inherit");
	     if (obj)
	     {
		 cats = find_cats_class (obj, 0, nil);
		 interp_result_list_ids (interp, cats);
		 free_list (cats);
	     }
	}
	else
	{
	    obj = lookup_required_name (clientData, interp, argv[2], "kb attributes");
	    if (! obj)
		return TCL_ERROR;

	    if (argc < 4)
	    {
		/* wants us to return list of attributes */
		interp_result_list_ids (interp, obj->attributes);
	    }
	    else if (argc < 6)
	    {
		interp->result = "Wrong # args for attr2";
		return TCL_ERROR;
	    }
	    else
	    {
		return process_attr_def2 (interp, obj, argc, argv);
	    }
	}
    }
    else if (streq (argv[1], "attr_of_cat"))
    {
	obj = lookup_required_name (clientData, interp, argv[2], "kb attr_of_cat");
	if (obj)
	{
	    Boolean hit;

	    for (ptr = obj->attributes; ptr; ptr = ptr->next)
	    {
		hit = true;
		for (i = 3; i < argc; i++)
		{
		    if (! (attr_inst_of (ptr->object, argv[i])))
		    {
			hit = false;
			break;
		    }
		}
		if (hit)
		    Tcl_AppendElement (interp, ptr->object->id);
	    }
	}
    }
    else if (arg_len >= 3 && strncmp (argv[1], "match_names", arg_len) == 0)
    {
	/* kb match_names pattern */
	if (argc != 3)
	{
	    interp->result = "Wrong number of arguments: kb match_names pattern";
	    return TCL_ERROR;
	}
	if (TCL_ERROR == kb_match_names (interp, argv[2]))
	    return TCL_ERROR;
    }
# ifdef CKB
    else if (arg_len >= 3 && strncmp (argv[1], "download", arg_len) == 0)
    {
	if (argc != 4)
	{
	    interp->result = "Wrong number of arguments: kb download ckb classname";
	    return TCL_ERROR;
	}
	if (! ckb_download (argv[3], argv[2], interp, 0, 10000, 1))
	    return TCL_ERROR;
    }
# endif
    else if (strncasecmp (argv[1], "min", 3) == 0)
    {
	ptr = nil;
	for (i = 2; i < argc; i++)
	{
	    obj = lookup_required_name (clientData, interp, argv[i], "kb minimize");
	    if (! obj)
		return TCL_ERROR;
	    ptr = cons (obj, ptr);
	}
	ptr = minimize_classes (ptr, true);
	interp_result_list_ids (interp, ptr);
	free_list (ptr);
    }
    else if (strncasecmp (argv[1], "max", 3) == 0)
    {
	ptr = nil;
	for (i = 2; i < argc; i++)
	{
	    obj = lookup_required_name (clientData, interp, argv[i], "kb maximize");
	    if (! obj)
		return TCL_ERROR;
	    ptr = cons (obj, ptr);
	}
	ptr = maximize_classes (ptr);
	interp_result_list_ids (interp, ptr);
	free_list (ptr);
    }
    else if (strcmp (argv[1], "tds") == 0)
    {
	Tcl_DString* arg_str;
	if (strncmp (argv[2], "-call", 4) == 0)
	{
	    ARGCHECK (5,0,"tds [-callback callback] command [arg1 arg2 ...]")
	    arg_str = compose_rep_args (argc - 5, argv + 5);
	    RepGeneric (clientData, interp, argv[4], Tcl_DStringValue (arg_str), argv[3]);
	}
	else
	{
	    ARGCHECK (3,0,"tds [-callback callback] command [arg1 arg2 ...]")
	    arg_str = compose_rep_args (argc - 3, argv + 3);
	    RepGeneric (clientData, interp, argv[2], Tcl_DStringValue (arg_str), nil);
	}
	Tcl_DStringResult (interp, &rep_result);
	if (rep_error_code != TCL_OK)
	    return rep_error_code;
    }
    else if (strncmp (argv[1], "-call", 4) == 0)
    {
	Tcl_DString* arg_str;
	ARGCHECK (3, 0, "-call call tds_command [arg1 arg2 ...]");
	if (get_tm_command (argv[3]))
	{
	    arg_str = compose_rep_args (argc - 4, argv + 4);
	    RepGeneric (clientData, interp, argv[3], Tcl_DStringValue (arg_str), argv[2]);
	    Tcl_DStringResult (interp, &rep_result);
	    if (rep_error_code != TCL_OK)
		return rep_error_code;
	}
	else
	{
	    Tcl_AppendResult (interp, "kb: bad tds command: ", argv[3], 0);
	    return TCL_ERROR;
	}
    }
    else if (get_tm_command (argv[1]))
    {
	Tcl_DString* arg_str;
	ARGCHECK (2,0,"tds_command [arg1 arg2 ...]");
	arg_str = compose_rep_args (argc - 2, argv + 2);
	RepGeneric (clientData, interp, argv[1], Tcl_DStringValue (arg_str), nil);
	Tcl_DStringResult (interp, &rep_result);
	if (rep_error_code != TCL_OK)
	    return rep_error_code;
    }
    else
    {
	interp->result = "Usage: kb <command> obj_name [val1 val2 ...]";
	return TCL_ERROR;
    }

    return TCL_OK;
}

KBObject* new_prim_object (name, attr_of, type)
    char* name;
    KBObject* attr_of;
    long type;
{
    KBObject* obj;

    obj = new_object (name, attr_of);

    SETLEVEL (obj, type);

    return obj;
}


void init_rule_kb ();

void init_kb ()
{
    KBObject* obj;
    current_source = new (Source);
    current_source->source_type = SOURCE_BUILTIN;

    initializing = true;
    debug_conn = 0;
    obj_id = 0;
    parser_new_objects = nil;
    num_allocated_objects = INIT_ALLOCATION;
    all_objects = (KBObject**) malloc (sizeof (KBObject*) * num_allocated_objects);

    Tcl_InitHashTable (&kb_symbols, TCL_STRING_KEYS);
    Tcl_DStringInit (&rep_result);
    Tcl_DStringInit (&compose_result_string);
    init_rule_kb ();
    /* fills the slot of object 0 which we cannot refer to from wish */

    new_object ("fadferaf98a7dfadf3473462346304*&^((**&", nil);
    proposition_class = new_prim_object (PROP, nil, L_OMEGA);
    class_class = new_prim_object ("Class", nil);
    add_isa_parent (proposition_class, class_class, L_OMEGA, false);
    add_instance_of (proposition_class, class_class, false);
    sclass_class = new_prim_object ("SimpleClass", nil, L_M1);
    add_isa_parent (sclass_class, class_class, false);
    add_instance_of (sclass_class, class_class, false);
    token_class = new_prim_object ("Token", nil, L_SCLASS);
    add_instance_of (token_class, sclass_class, false);
    m1class_class = new_prim_object ("MetaClass", nil, L_M2);
    add_isa_parent (m1class_class, class_class, false);
    add_instance_of (m1class_class, class_class, false);
    m2class_class = new_prim_object ("MetaMetaClass", nil, L_M3);
    add_isa_parent (m2class_class, class_class, false);
    add_instance_of (m2class_class, class_class, false);
    m3class_class = new_prim_object ("MetaMetaMetaClass", nil, L_M4);
    add_isa_parent (m3class_class, class_class, false);
    add_instance_of (m3class_class, class_class, false);
    m4class_class = new_prim_object ("MetaMetaMetaMetaClass", nil, 0);
    add_isa_parent (m4class_class, class_class, false);
    add_instance_of (m4class_class, class_class, false);
    attribute_class = new_prim_object ("attribute", class_class, L_ATTRIBUTE);
    add_instance_of (attribute_class, attribute_class, false);
    add_isa_parent (attribute_class, class_class, false);
    push (attribute_class, class_class->attributes);
    attribute_class->value = class_class;

    add_instance_of (sclass_class, m1class_class, false);
    add_instance_of (m1class_class, m2class_class, false);
    add_instance_of (m2class_class, m3class_class, false);
    add_instance_of (m3class_class, m4class_class, false);



    initializing = false;

    pstring_class = new_prim_object ("PString", nil, L_SCLASS);
    add_instance_of (pstring_class, sclass_class, false);

    string_class = new_prim_object ("String", nil, L_SCLASS);
    add_instance_of (string_class, sclass_class, false);

    obj = new_prim_object ("Integer", nil, L_SCLASS);
    add_instance_of (obj, sclass_class, false);

    number_class = new_prim_object ("Number", nil, L_SCLASS);
    add_instance_of (number_class, sclass_class, false);
    add_isa_parent (obj, number_class, false);

    dump_start = obj_id;
    tm_init ();

    /* create the default source */
    default_source = new (Source);
    default_source->source_type = SOURCE_TCL;
    current_source = default_source;
}

# ifdef CKB
   extern int ckb_exit ();
# endif

void registerKBCommands (interp)
    Tcl_Interp* interp;
{

    Tcl_CreateCommand (interp, "kb", kb_comm, (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
# ifdef CKB
    Tcl_CreateCommand (interp, "exit", ckb_exit, (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
    Tcl_SetVar (interp, "versant_enabled", "1", TCL_GLOBAL_ONLY);
    confirm_interp = Tcl_CreateInterp();
# endif
    init_kb ();
}


Tcl_HashTable kb_symbols;

void register_tm_command (name, command)
    char* name;
    void* command;
{
    Tcl_HashEntry* entry;
    int newp;

    entry = Tcl_CreateHashEntry (&kb_symbols, name, &newp);
    Tcl_SetHashValue (entry, command);
}


void* get_tm_command (name)
    char* name;
{
    Tcl_HashEntry* entry;

    entry = Tcl_FindHashEntry (&kb_symbols, name);
    if (entry)
	return (void*) Tcl_GetHashValue (entry);
    else
	return nil;
}

char rep_wait_var[] = "waiting  for rep to return";
char* rep_wait_args[] =
{
    "wait",
    "variable",
    "waiting  for rep to return"
};



void rep_wait (clientData, interp)
    ClientData clientData;
    Tcl_Interp* interp;
{
    char* val;

    val = Tcl_GetVar (interp, rep_wait_var, TCL_GLOBAL_ONLY);
    if (! val || ! streq (val, "W"))
    {
	rep_error_code = TCL_OK;
	Tcl_DStringFree (&rep_result);
	Tcl_SetVar (interp, rep_wait_var, "W", TCL_GLOBAL_ONLY);
	Tk_TkwaitCmd (clientData, interp, 3, rep_wait_args);
    }
}

/* WINGS BEGIN */
#ifdef WINGS_UNIX
void rep_end_wait ()
{
    Tcl_SetVar (global_interp, rep_wait_var, "1", TCL_GLOBAL_ONLY);
}
#endif
/* WINGS END */

void rep_append_result (interp, x)
    char* x;
{
    Tcl_DStringAppendElement (&rep_result, x);
}

void rep_append_result_id (interp, x)
    char* x;
{
    KBObject* obj;

    obj = new_object (x, nil);
    Tcl_DStringAppendElement (&rep_result, obj->id);
}

/* WINGS BEGIN */
#ifdef WINGS_UNIX
void rep_error (message1, message2)
    char* message1;
    char* message2;
{
    rep_error_code = TCL_ERROR;
    Tcl_DStringFree (&rep_result);

    if (message1)
	rep_append_result (global_interp, message1);

    if (message2)
	rep_append_result (global_interp, message2);
}
#endif
/* WINGS END */

void kb_update_status (clientData, interp, names, definep)
    ClientData clientData;
    Tcl_Interp* interp;
    char*	names;
    Boolean	definep;
{
    char** namev;
    int    name_count;
    KBObject* obj;
    int i;

    if (names && (TCL_OK == Tcl_SplitList (interp, names, &name_count, &namev)))
    {
	for (i = 0; i < name_count; i++)
	{
	    obj = get_object (namev[i]);
	    if (obj)
		obj->status &= ~TDS_REQUESTED;
	}
/* WINGS BEGIN */
/*	free (namev); */
/* WINGS END */
    }
}


void rep_set_tds_error (name)
    char* name;
{
    KBObject* obj;

    obj = get_object (name);
    if (obj)
    {
	obj->status |= TDS_REQUESTED | TDS_UNKNOWN;
    }
}


static KBList* child_list;

void reset_child_list ()
{
    child_list = nil;
}

void rep_add_child (name)
    char* name;
{
    push (new_object (name, nil), child_list);
}

long lower_level (level)
    long level;
{
    switch (level)
    {
      case L_OMEGA:
	return L_UNKNOWN;

      case L_TOKEN:
	return L_UNKNOWN;

      case L_SCLASS:
	return L_TOKEN;

      case L_M1:
	return L_SCLASS;

      case L_M2:
	return L_M1;

      case L_M3:
	return L_M2;

      case L_M4:
	return L_M3;

      default:
	return L_UNKNOWN;
    }
}

void rep_set_children (clientData, interp, names)
    ClientData clientData;
    Tcl_Interp* interp;
    char*	names;
{
    char** namev;
    int    name_count;
    KBList* ptr;
    KBObject* obj;
    long   level;

    if (names && (TCL_OK == Tcl_SplitList (interp, names, &name_count, &namev)))
    {
	if (name_count > 0)
	{
	    obj = get_object (namev[0]);
	    if (debug_conn)
		fprintf (stderr, "RSC %s 0x%x", namev[0], obj);
	    if (obj)
	    {
		level = get_level (obj);
		obj->status |= TDS_CHILDREN_RECVD;
		for (ptr = child_list; ptr; ptr = ptr->next)
		{
		    if (debug_conn)
			fprintf (stderr, " %s %x",
				 ptr->object->name ? ptr->object->name : ptr->object->id,
				 level);
		    push (ptr->object, obj->children);
		    /* assume kb is valid */
		    if (! LEVELOF (ptr->object))
			SETLEVEL (ptr->object, level);
		}
	    }
	    free_list (child_list);
	    child_list = nil;
	    if (debug_conn)
		fprintf (stderr, "\n");
	}
/* WINGS BEGIN */
/*	free (namev); */
/* WINGS END */
    }
}

void rep_set_instances (clientData, interp, names)
    ClientData clientData;
    Tcl_Interp* interp;
    char*	names;
{
    char** namev;
    int    name_count;
    KBList* ptr;
    KBObject* obj;
    long   level;

    if (names && (TCL_OK == Tcl_SplitList (interp, names, &name_count, &namev)))
    {
	if (name_count > 0)
	{
	    obj = get_object (namev[0]);
	    if (debug_conn)
		fprintf (stderr, "RSI %s 0x%x", namev[0], obj);
	    if (obj)
	    {
		level = lower_level (get_level (obj));
		obj->status |= TDS_INST_RECVD;
		for (ptr = child_list; ptr; ptr = ptr->next)
		{
		    if (debug_conn)
			fprintf (stderr, " %s %x",
				 ptr->object->name ? ptr->object->name : ptr->object->id,
				 level);
		    push (ptr->object, obj->instances);
		    /* assume kb is valid */
		    if (! LEVELOF (ptr->object))
			SETLEVEL (ptr->object, level);
		}
	    }
	    free_list (child_list);
	    child_list = nil;
	    if (debug_conn)
		fprintf (stderr, "\n");
	}
/* WINGS BEGIN */
/*	free (namev); */
/* WINGS END */
    }
}



void kb_call_back (clientData, interp, callback, errorp, answer)
    ClientData clientData;
    Tcl_Interp* interp;
    char* callback;
    int errorp;
    char* answer;
{
    char* errorp_field;
    int result;

    if (errorp)
	errorp_field = " 1 ";
    else
	errorp_field = " 0 ";

    result = Tcl_VarEval (interp, callback, errorp_field, " {", answer,  "}", 0);
    if (result != TCL_OK)
    {
	Tcl_VarEval (interp, "tkerror {", interp->result, "}", 0);
    }
}

static Source* save_ckb_source;

void source_enter_ckb (kb)
    char* kb;
{
    save_ckb_source = current_source;

    current_source = new (Source);
    current_source->source_type = SOURCE_CKB;
    current_source->source.ckb.name = strdup (kb);
}

void source_exit_ckb ()
{
    current_source = save_ckb_source;
}
