
# line 2 "text-parser.y"
/*
 * text-parser.y --
 *
 *  Last Modified: Mon Sep 18 10:02:13 1995 by Bryan M. Kramer
 *
 * Copyright (c) 1990-1994 University of Toronto.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF Toronto BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * Toronto HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF Toronto SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF Toronto HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

# include <stdio.h>
# include <stdlib.h>
# include "kb.h"
# include "dr.h"

# define yywrap() 1

typedef union {
    int ival;
    float fval;
    void *sval;
    char* strval;
    KBObject* objval;
    KBList* listval;
    Atom*   atomval;
    Term*   termval;
    Variable* varval;
} YYSTYPE;

KBObject* current_object;
KBList*  categories;
KBList*	 current_sequence;
char*    current_thing;

extern VarList*    rule_variables;
extern int	   num_rule_variables;
extern VarList*    last_rule_variable;


void attribute_from_current ();
void attribute_from_object ();
void attribute_from_sequence ();

# define TOKEN_KEYWORD "Token"
# define SCLASS_KEYWORD "SClass"
# define M1CLASS_KEYWORD "M1Class"
# define M2CLASS_KEYWORD "M2Class"

#undef YYMAXDEPTH
#define YYMAXDEPTH 10000
#undef YYERROR_VERBOSE
# define TP_RULE 257
# define TP_NOT 258
# define TP_GOAL 259
# define TP_TOKEN 260
# define TP_SCLASS 261
# define TP_M1CLASS 262
# define TP_M2CLASS 263
# define TP_M3CLASS 264
# define TP_IN 265
# define TP_ISA 266
# define TP_WITH 267
# define TP_END 268
# define TP_COMMA 269
# define TP_COLON 270
# define TP_SLASH 271
# define TP_DOLLAR 272
# define TP_EQUAL 273
# define TP_nil 274
# define TP_SEMICOLON 275
# define INSTANCE 276
# define SLOT 277
# define LBRACKET 278
# define RBRACKET 279
# define LBRACE 280
# define RBRACE 281
# define START_SEQ 282
# define END_SEQ 283
# define TP_IDENTIFIER 284
# define TP_STRING 285
# define TP_OBJECT 286
# define TP_DASH 287

#ifdef __STDC__
#include <stdlib.h>
#include <string.h>
#else
#include <malloc.h>
#include <memory.h>
#endif

#include <values.h>

#ifdef __cplusplus

#ifndef yyerror
	void yyerror(const char *);
#endif

#ifndef yylex
#ifdef __EXTERN_C__
	extern "C" { int yylex(void); }
#else
	int yylex(void);
#endif
#endif
	int yyparse(void);

#endif
#define yyclearin yychar = -1
#define yyerrok yyerrflag = 0
extern int yychar;
extern int yyerrflag;
YYSTYPE yylval;
YYSTYPE yyval;
typedef int yytabelem;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif
#if YYMAXDEPTH > 0
int yy_yys[YYMAXDEPTH], *yys = yy_yys;
YYSTYPE yy_yyv[YYMAXDEPTH], *yyv = yy_yyv;
#else	/* user does initial allocation */
int *yys;
YYSTYPE *yyv;
#endif
static int yymaxdepth = YYMAXDEPTH;
# define YYERRCODE 256

# line 394 "text-parser.y"

/* Programs */


int use_string;
int string_pos;
int string_len;
char* input_string;


#include "scanner.h"


int parse_error;
char* cur_file;

void yyerror (s)
    char	*s;
{
    parse_error = 1;
    if (use_string)
	fprintf (stderr, "%s at position %d in string '%s' near %s\n", s, string_pos,
		 input_string, yytext);
    else
	fprintf(stderr, "Syntax error on line %d of %s near %s\n",
		yylineno, cur_file, yytext);
}


void semantic_msg (s)
    char	*s;
{
    parse_error = 1;
    if (use_string)
	fprintf (stderr, "%s at %d in string '%s' near %s\n",
		 s, string_pos, input_string, yytext);
    else
	fprintf (stderr, "%s on line %d of %s near %s\n", s, yylineno, cur_file, yytext);
}



int parse_string (str)
    char* str;
{
    use_string = 1;
    string_pos = 0;
    string_len = strlen (str);
    input_string = str;
    reset_new_object_list ();
    yyparse ();
    return parse_error;
}



int parse_file (file)
    char* file;
{
    Source* save_source;
    save_source = current_source;
    current_source = new (Source);
    current_source->source_type = SOURCE_FILE;
    current_source->source.file.name = strdup (file);

    cur_file = file;
    use_string = 0;
    current_object = nil;
    categories = nil;
    parse_error = 0;
    yylineno = 1;
    yyin = fopen (cur_file, "r");
    if (! yyin)
    {
	fprintf (stderr, "Couldn't open file %s\n", cur_file);
	return;
    }
    reset_new_object_list ();
    yyparse ();

    fclose (yyin);

    current_source = save_source;

    return parse_error;
}


/* current thing and current sequence are mutually exclusive */

void attribute_from_current ()
{
    KBList* ptr;
    if (current_thing)
    {
	create_attr (current_object, nil, categories, new_object (current_thing, nil));
    }
    else
    {
	for (ptr = current_sequence; ptr; ptr = ptr->next)
	    create_attr (current_object, nil, categories, ptr->object);

	free_list (current_sequence);
    }
}


void attribute_from_object (object)
    KBObject* object;
{
    KBList* ptr;
    if (current_thing)
    {
	create_attr (current_object, current_thing, categories, object);
    }
    else
    {
	for (ptr = current_sequence; ptr; ptr = ptr->next)
	    create_attr (current_object, ptr->object->name, categories, object);

	free_list (current_sequence);
    }
}



void attribute_from_sequence (seq)
    KBList* seq;
{
    KBList* ptr;
    KBList* ptr2;
    if (current_thing)
    {
	for (ptr = seq; ptr; ptr = ptr->next)
	    create_attr (current_object, current_thing, categories, ptr->object);
    }
    else
    {
	for (ptr = seq; ptr; ptr = ptr->next)
	{
	    for (ptr2 = current_sequence; ptr; ptr = ptr->next)
		create_attr (current_object, ptr->object->name, categories, ptr2->object);
	}

	free_list (current_sequence);
    }

    free_list (seq);
}
yytabelem yyexca[] ={
-1, 1,
	0, -1,
	-2, 0,
	};
# define YYNPROD 115
# define YYLAST 311
yytabelem yyact[]={

    40,    42,    41,     9,    10,    11,    12,    53,    54,    55,
   103,    26,   120,    81,   174,    98,    97,   125,   178,   151,
   173,   153,   147,    95,   102,   148,    72,    36,    39,    38,
    40,    42,    41,     9,    10,    11,    12,    40,    42,    41,
     9,    10,    11,    12,   161,   163,   162,     9,    10,    11,
    12,   172,    26,   145,   142,   122,   110,    36,    39,    38,
    91,    89,   164,    43,    36,    39,    38,   109,   138,   164,
   124,   157,   160,   159,    40,    42,    41,     9,    10,    11,
    12,    40,    42,    41,     9,    10,    11,    12,    99,    86,
    22,   133,    23,     9,    10,    11,    12,     9,    10,    11,
    12,    36,    39,    38,     9,    10,    11,    12,    36,    39,
    38,   128,    49,    65,    45,    29,   135,    26,   118,   105,
     8,    26,     5,    83,    63,    76,    60,    34,    16,    94,
   146,   170,    24,   156,   149,   108,   101,   134,    37,     7,
    30,    31,     7,   104,    25,    35,   113,    92,    82,   127,
    87,    62,    46,    66,    78,    51,    47,    27,    21,    80,
    25,     2,    56,    61,    79,    13,    77,    59,   166,   167,
    25,   154,   139,   137,   123,    67,    28,    33,    15,    75,
    50,    58,    71,    32,    14,     6,    85,    64,    44,    18,
    57,    17,    19,    84,    20,    88,     4,    67,     3,    96,
    93,    70,    48,    68,   155,    69,    52,   131,    74,    73,
     1,     0,     0,   106,     0,     0,     0,     0,    88,     0,
     0,    90,     0,     0,     0,   114,     0,     0,     0,     0,
   115,   116,   117,   119,   100,     0,   107,     0,     0,   111,
     0,   112,     0,     0,   114,     0,     0,     0,   132,     0,
     0,     0,     0,     0,     0,   121,     0,   140,   114,     0,
     0,   126,     0,     0,   136,   130,   129,     0,     0,   143,
     0,     0,   144,   150,     0,   141,     0,     0,     0,     0,
     0,   152,     0,   150,   168,   165,   158,     0,   169,     0,
     0,     0,     0,     0,   171,     0,     0,     0,     0,     0,
     0,   171,   176,   175,   177,     0,     0,     0,   171,     0,
   179 };
yytabelem yypact[]={

  -156,-10000000,-10000000,  -156,  -137,  -167,  -162,  -273,  -273,-10000000,
-10000000,-10000000,-10000000,-10000000,  -139,  -176,-10000000,  -216,  -164,  -273,
  -166,  -163,-10000000,-10000000,-10000000,-10000000,-10000000,-10000000,  -277,  -273,
-10000000,  -163,  -141,  -176,-10000000,  -145,-10000000,-10000000,-10000000,-10000000,
-10000000,-10000000,-10000000,-10000000,  -165,  -176,  -166,-10000000,  -166,  -232,
-10000000,  -162,  -277,-10000000,-10000000,-10000000,-10000000,-10000000,  -143,  -271,
-10000000,  -146,-10000000,  -176,  -189,  -176,  -218,  -176,-10000000,-10000000,
  -219,  -257,  -190,-10000000,-10000000,-10000000,-10000000,-10000000,-10000000,  -271,
  -260,  -150,-10000000,  -176,  -145,-10000000,  -211,  -223,  -176,-10000000,
-10000000,-10000000,-10000000,  -257,-10000000,  -176,-10000000,  -273,  -273,  -273,
-10000000,-10000000,  -152,  -176,-10000000,  -272,  -146,-10000000,  -224,  -208,
-10000000,-10000000,-10000000,  -264,  -176,  -160,  -160,  -257,  -183,  -159,
  -150,-10000000,-10000000,  -210,  -273,-10000000,-10000000,-10000000,  -176,-10000000,
  -225,  -159,-10000000,-10000000,-10000000,  -260,-10000000,  -226,  -256,  -254,
  -273,  -252,-10000000,-10000000,-10000000,  -211,  -258,  -213,-10000000,-10000000,
  -273,-10000000,-10000000,-10000000,  -220,-10000000,-10000000,-10000000,-10000000,-10000000,
-10000000,-10000000,-10000000,-10000000,  -227,-10000000,-10000000,  -259,-10000000,-10000000,
  -269,  -227,  -176,  -256,-10000000,-10000000,  -261,-10000000,  -227,-10000000 };
yytabelem yypgo[]={

     0,   210,   138,   158,   132,   129,   207,   206,   204,   131,
   133,   156,   147,   149,   146,   202,   201,   200,   199,   161,
   198,   196,   194,   192,   191,   189,   188,   187,   186,   185,
   184,   183,   181,   179,   157,   153,   150,   178,   151,   177,
   148,   176,   155,   135,   174,   173,   172,   134,   171,   169,
   168,   130,   167,   166,   154,   164,   159,   136,   143,   137 };
yytabelem yyr1[]={

     0,     1,    19,    19,    21,     3,     3,    22,    23,    24,
    24,    24,    11,    11,    16,    16,    15,    12,    12,    17,
    17,    17,    18,    18,    13,    13,    14,    14,    25,    29,
    20,    20,    20,     2,     2,     2,     2,     4,    26,    35,
    35,    27,    36,    36,    30,    30,    37,    38,    38,    31,
    31,    39,    40,    40,    41,     7,     7,     7,    42,    42,
    34,    34,    28,    43,    43,    44,    47,    47,    46,    46,
     9,     9,     9,    10,    48,    48,    49,    49,    49,    50,
    51,    51,    45,    45,    32,    32,    52,    53,    54,    54,
    55,    56,    58,    58,     6,     6,    57,    57,    59,    59,
     5,     5,     5,     5,     5,     5,     5,     8,     8,     8,
     8,     8,     8,     8,    33 };
yytabelem yyr2[]={

     0,     3,     1,     5,     5,     3,     3,     3,     3,     8,
     7,     5,     3,     5,     5,    11,     7,     1,     5,     3,
     7,     3,     7,     7,     1,     7,     1,     5,     5,     7,
    11,     6,     4,     3,     3,     3,     3,     3,     6,     0,
     5,     7,     0,     5,     0,     7,     2,     1,     7,     0,
     7,     2,     1,     7,     5,     3,     3,     3,     0,     5,
     0,     6,     6,     0,    10,     6,     0,     5,     0,     5,
     1,     5,     9,     7,     3,     3,     1,     3,     3,     6,
     0,     7,     0,     6,     0,     4,     3,     3,     1,     5,
     4,     5,     1,     7,     3,     3,     9,     7,     1,     5,
     3,     3,     3,     3,     3,     3,     3,     3,     3,     3,
     3,     3,     3,     3,     3 };
yytabelem yychk[]={

-10000000,    -1,   -19,   -20,   -21,   278,   -29,    -2,   276,   260,
   261,   262,   263,   -19,   -30,   -37,   265,   -24,   -25,   -23,
   -22,    -3,   257,   259,    -4,    -2,   284,   -34,   -41,   277,
    -4,    -4,   -31,   -39,   266,    -5,   284,    -2,   286,   285,
   257,   259,   258,   279,   -26,   278,    -4,   -11,   -15,   278,
    -3,   -42,    -7,   284,   285,   286,    -4,    -3,   -32,   -52,
   267,    -5,   -38,   269,   -27,   278,   -35,    -5,   -11,   -11,
   -16,    -4,   258,   -34,   -42,   -33,   268,   -53,   -54,   -55,
   -56,   284,   -40,   269,    -5,   -28,   278,   -36,    -5,   279,
   -35,   279,   -12,   -17,    -5,   280,   -18,   273,   272,   278,
   -54,   -57,   284,   270,   -58,   269,    -5,   -38,   -43,   278,
   279,   -36,   -12,   -14,    -5,    -4,    -4,    -4,   270,    -5,
   284,   -40,   279,   -44,   278,   281,   -14,   -13,   271,   -13,
   -12,    -6,    -5,   274,   -59,   275,   -58,   -45,   278,   -46,
    -4,   -14,   279,   -59,   -57,   279,   -51,   278,   279,   -47,
    -4,   271,   -43,   279,   -48,    -8,   -10,   284,    -2,   286,
   285,   257,   259,   258,   282,   -47,   -50,   -49,    -5,   -10,
    -9,    -5,   278,   279,   283,    -9,    -5,   -51,   279,    -9 };
yytabelem yydef[]={

     2,    -2,     1,     2,    44,     0,    60,     0,     0,    33,
    34,    35,    36,     3,    49,     0,    46,     0,     0,     0,
     0,     0,     8,     7,     5,     6,    37,    32,    58,     0,
     4,     0,    84,     0,    51,    47,   100,   101,   102,   103,
   104,   105,   106,    31,     0,    39,     0,    11,    12,     0,
    28,    60,    58,    55,    56,    57,    54,    29,     0,    88,
    86,    52,    45,     0,     0,    42,     0,    39,    10,    13,
     0,    17,     0,    61,    59,    30,   114,    85,    87,    88,
     0,    92,    50,     0,    47,     9,    63,     0,    42,    38,
    40,    16,    14,    17,    19,    26,    21,     0,     0,     0,
    89,    90,     0,     0,    91,     0,    52,    48,     0,     0,
    41,    43,    18,     0,    26,    24,    24,    17,     0,    98,
    92,    53,    62,    82,    68,    20,    27,    22,    26,    23,
     0,    98,    94,    95,    97,     0,    93,     0,    80,     0,
    66,     0,    15,    96,    99,    63,     0,     0,    65,    69,
    66,    25,    64,    83,    76,    74,    75,   107,   108,   109,
   110,   111,   112,   113,    70,    67,    81,     0,    77,    78,
     0,    70,     0,    80,    73,    71,     0,    79,    70,    72 };
typedef struct
#ifdef __cplusplus
	yytoktype
#endif
{ char *t_name; int t_val; } yytoktype;
#ifndef YYDEBUG
#	define YYDEBUG	0	/* don't allow debugging */
#endif

#if YYDEBUG

yytoktype yytoks[] =
{
	"TP_RULE",	257,
	"TP_NOT",	258,
	"TP_GOAL",	259,
	"TP_TOKEN",	260,
	"TP_SCLASS",	261,
	"TP_M1CLASS",	262,
	"TP_M2CLASS",	263,
	"TP_M3CLASS",	264,
	"TP_IN",	265,
	"TP_ISA",	266,
	"TP_WITH",	267,
	"TP_END",	268,
	"TP_COMMA",	269,
	"TP_COLON",	270,
	"TP_SLASH",	271,
	"TP_DOLLAR",	272,
	"TP_EQUAL",	273,
	"TP_nil",	274,
	"TP_SEMICOLON",	275,
	"INSTANCE",	276,
	"SLOT",	277,
	"LBRACKET",	278,
	"RBRACKET",	279,
	"LBRACE",	280,
	"RBRACE",	281,
	"START_SEQ",	282,
	"END_SEQ",	283,
	"TP_IDENTIFIER",	284,
	"TP_STRING",	285,
	"TP_OBJECT",	286,
	"TP_DASH",	287,
	"-unknown-",	-1	/* ends search */
};

char * yyreds[] =
{
	"-no such reduction-",
	"file : declarations",
	"declarations : /* empty */",
	"declarations : declaration declarations",
	"objDecl : levelKeyword name",
	"cname : name",
	"cname : levelKeyword",
	"goalStart : TP_GOAL",
	"ruleStart : TP_RULE",
	"listDecl : instanceHead instanceList isaList withList",
	"listDecl : ruleStart name atoms",
	"listDecl : goalStart atoms",
	"atoms : one_atom",
	"atoms : one_atom atoms",
	"atom_kind : name terms",
	"atom_kind : TP_NOT LBRACKET name terms RBRACKET",
	"one_atom : LBRACKET atom_kind RBRACKET",
	"terms : /* empty */",
	"terms : one_term terms",
	"one_term : ObjectID",
	"one_term : LBRACE term_objects RBRACE",
	"one_term : variable",
	"variable : TP_EQUAL name var_types",
	"variable : TP_DOLLAR name var_types",
	"var_types : /* empty */",
	"var_types : TP_SLASH term_objects TP_SLASH",
	"term_objects : /* empty */",
	"term_objects : ObjectID term_objects",
	"instanceHead : cname cname",
	"instanceDecl : INSTANCE name cname",
	"declaration : objDecl inClause isaClause withClause endKeyword",
	"declaration : LBRACKET listDecl RBRACKET",
	"declaration : instanceDecl instanceSlots",
	"levelKeyword : TP_TOKEN",
	"levelKeyword : TP_SCLASS",
	"levelKeyword : TP_M1CLASS",
	"levelKeyword : TP_M2CLASS",
	"name : TP_IDENTIFIER",
	"instanceList : LBRACKET instanceAtom RBRACKET",
	"instanceAtom : /* empty */",
	"instanceAtom : ObjectID instanceAtom",
	"isaList : LBRACKET isaAtom RBRACKET",
	"isaAtom : /* empty */",
	"isaAtom : ObjectID isaAtom",
	"inClause : /* empty */",
	"inClause : inKeyword ObjectID inClauseIDs",
	"inKeyword : TP_IN",
	"inClauseIDs : /* empty */",
	"inClauseIDs : TP_COMMA ObjectID inClauseIDs",
	"isaClause : /* empty */",
	"isaClause : isaKeyword ObjectID isaClauseIDs",
	"isaKeyword : TP_ISA",
	"isaClauseIDs : /* empty */",
	"isaClauseIDs : TP_COMMA ObjectID isaClauseIDs",
	"instanceSlotStart : SLOT name",
	"instanceSlotValueValue : TP_IDENTIFIER",
	"instanceSlotValueValue : TP_STRING",
	"instanceSlotValueValue : TP_OBJECT",
	"instanceSlotValues : /* empty */",
	"instanceSlotValues : instanceSlotValueValue instanceSlotValues",
	"instanceSlots : /* empty */",
	"instanceSlots : instanceSlotStart instanceSlotValues instanceSlots",
	"withList : LBRACKET withGroup RBRACKET",
	"withGroup : /* empty */",
	"withGroup : LBRACKET withGroupCats withGroupAttrs RBRACKET withGroup",
	"withGroupCats : LBRACKET withGroupCats1 RBRACKET",
	"withGroupCats2 : /* empty */",
	"withGroupCats2 : name withGroupCats2",
	"withGroupCats1 : /* empty */",
	"withGroupCats1 : name withGroupCats2",
	"objects : /* empty */",
	"objects : ObjectID objects",
	"objects : LBRACKET ObjectID RBRACKET objects",
	"sequence : START_SEQ objects END_SEQ",
	"withGroupAttrs2Head : thing",
	"withGroupAttrs2Head : sequence",
	"withGroupDoIt : /* empty */",
	"withGroupDoIt : ObjectID",
	"withGroupDoIt : sequence",
	"withGroupAttrs2Tail : withGroupDoIt RBRACKET withGroupAttrs2",
	"withGroupAttrs2 : /* empty */",
	"withGroupAttrs2 : LBRACKET withGroupAttrs2Head withGroupAttrs2Tail",
	"withGroupAttrs : /* empty */",
	"withGroupAttrs : LBRACKET withGroupAttrs2 RBRACKET",
	"withClause : /* empty */",
	"withClause : withKeyword attrDecls",
	"withKeyword : TP_WITH",
	"attrDecls : attrDecls2",
	"attrDecls2 : /* empty */",
	"attrDecls2 : attrDecl attrDecls2",
	"attrDecl : categories valuePairs",
	"categories : TP_IDENTIFIER categories2",
	"categories2 : /* empty */",
	"categories2 : TP_COMMA TP_IDENTIFIER categories2",
	"optObjectID : ObjectID",
	"optObjectID : TP_nil",
	"valuePairs : TP_IDENTIFIER TP_COLON optObjectID valuePairs2",
	"valuePairs : TP_COLON ObjectID valuePairs2",
	"valuePairs2 : /* empty */",
	"valuePairs2 : TP_SEMICOLON valuePairs",
	"ObjectID : TP_IDENTIFIER",
	"ObjectID : levelKeyword",
	"ObjectID : TP_OBJECT",
	"ObjectID : TP_STRING",
	"ObjectID : TP_RULE",
	"ObjectID : TP_GOAL",
	"ObjectID : TP_NOT",
	"thing : TP_IDENTIFIER",
	"thing : levelKeyword",
	"thing : TP_OBJECT",
	"thing : TP_STRING",
	"thing : TP_RULE",
	"thing : TP_GOAL",
	"thing : TP_NOT",
	"endKeyword : TP_END",
};
#endif /* YYDEBUG */
/*
 * Copyright (c) 1993 by Sun Microsystems, Inc.
 */

#pragma ident	"@(#)yaccpar	6.12	93/06/07 SMI"

/*
** Skeleton parser driver for yacc output
*/

/*
** yacc user known macros and defines
*/
#define YYERROR		goto yyerrlab
#define YYACCEPT	return(0)
#define YYABORT		return(1)
#define YYBACKUP( newtoken, newvalue )\
{\
	if ( yychar >= 0 || ( yyr2[ yytmp ] >> 1 ) != 1 )\
	{\
		yyerror( "syntax error - cannot backup" );\
		goto yyerrlab;\
	}\
	yychar = newtoken;\
	yystate = *yyps;\
	yylval = newvalue;\
	goto yynewstate;\
}
#define YYRECOVERING()	(!!yyerrflag)
#define YYNEW(type)	malloc(sizeof(type) * yynewmax)
#define YYCOPY(to, from, type) \
	(type *) memcpy(to, (char *) from, yynewmax * sizeof(type))
#define YYENLARGE( from, type) \
	(type *) realloc((char *) from, yynewmax * sizeof(type))
#ifndef YYDEBUG
#	define YYDEBUG	1	/* make debugging available */
#endif

/*
** user known globals
*/
int yydebug;			/* set to 1 to get debugging */

/*
** driver internal defines
*/
#define YYFLAG		(-10000000)

/*
** global variables used by the parser
*/
YYSTYPE *yypv;			/* top of value stack */
int *yyps;			/* top of state stack */

int yystate;			/* current state */
int yytmp;			/* extra var (lasts between blocks) */

int yynerrs;			/* number of errors */
int yyerrflag;			/* error recovery flag */
int yychar;			/* current input token number */



#ifdef YYNMBCHARS
#define YYLEX()		yycvtok(yylex())
/*
** yycvtok - return a token if i is a wchar_t value that exceeds 255.
**	If i<255, i itself is the token.  If i>255 but the neither 
**	of the 30th or 31st bit is on, i is already a token.
*/
#if defined(__STDC__) || defined(__cplusplus)
int yycvtok(int i)
#else
int yycvtok(i) int i;
#endif
{
	int first = 0;
	int last = YYNMBCHARS - 1;
	int mid;
	wchar_t j;

	if(i&0x60000000){/*Must convert to a token. */
		if( yymbchars[last].character < i ){
			return i;/*Giving up*/
		}
		while ((last>=first)&&(first>=0)) {/*Binary search loop*/
			mid = (first+last)/2;
			j = yymbchars[mid].character;
			if( j==i ){/*Found*/ 
				return yymbchars[mid].tvalue;
			}else if( j<i ){
				first = mid + 1;
			}else{
				last = mid -1;
			}
		}
		/*No entry in the table.*/
		return i;/* Giving up.*/
	}else{/* i is already a token. */
		return i;
	}
}
#else/*!YYNMBCHARS*/
#define YYLEX()		yylex()
#endif/*!YYNMBCHARS*/

/*
** yyparse - return 0 if worked, 1 if syntax error not recovered from
*/
#if defined(__STDC__) || defined(__cplusplus)
int yyparse(void)
#else
int yyparse()
#endif
{
	register YYSTYPE *yypvt;	/* top of value stack for $vars */

#if defined(__cplusplus) || defined(lint)
/*
	hacks to please C++ and lint - goto's inside switch should never be
	executed; yypvt is set to 0 to avoid "used before set" warning.
*/
	static int __yaccpar_lint_hack__ = 0;
	switch (__yaccpar_lint_hack__)
	{
		case 1: goto yyerrlab;
		case 2: goto yynewstate;
	}
	yypvt = 0;
#endif

	/*
	** Initialize externals - yyparse may be called more than once
	*/
	yypv = &yyv[-1];
	yyps = &yys[-1];
	yystate = 0;
	yytmp = 0;
	yynerrs = 0;
	yyerrflag = 0;
	yychar = -1;

#if YYMAXDEPTH <= 0
	if (yymaxdepth <= 0)
	{
		if ((yymaxdepth = YYEXPAND(0)) <= 0)
		{
			yyerror("yacc initialization error");
			YYABORT;
		}
	}
#endif

	{
		register YYSTYPE *yy_pv;	/* top of value stack */
		register int *yy_ps;		/* top of state stack */
		register int yy_state;		/* current state */
		register int  yy_n;		/* internal state number info */
	goto yystack;	/* moved from 6 lines above to here to please C++ */

		/*
		** get globals into registers.
		** branch to here only if YYBACKUP was called.
		*/
	yynewstate:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;
		goto yy_newstate;

		/*
		** get globals into registers.
		** either we just started, or we just finished a reduction
		*/
	yystack:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;

		/*
		** top of for (;;) loop while no reductions done
		*/
	yy_stack:
		/*
		** put a state and value onto the stacks
		*/
#if YYDEBUG
		/*
		** if debugging, look up token value in list of value vs.
		** name pairs.  0 and negative (-1) are special values.
		** Note: linear search is used since time is not a real
		** consideration while debugging.
		*/
		if ( yydebug )
		{
			register int yy_i;

			printf( "State %d, token ", yy_state );
			if ( yychar == 0 )
				printf( "end-of-file\n" );
			else if ( yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ++yy_ps >= &yys[ yymaxdepth ] )	/* room on stack? */
		{
			/*
			** reallocate and recover.  Note that pointers
			** have to be reset, or bad things will happen
			*/
			int yyps_index = (yy_ps - yys);
			int yypv_index = (yy_pv - yyv);
			int yypvt_index = (yypvt - yyv);
			int yynewmax;
#ifdef YYEXPAND
			yynewmax = YYEXPAND(yymaxdepth);
#else
			yynewmax = 2 * yymaxdepth;	/* double table size */
			if (yymaxdepth == YYMAXDEPTH)	/* first time growth */
			{
				char *newyys = (char *)YYNEW(int);
				char *newyyv = (char *)YYNEW(YYSTYPE);
				if (newyys != 0 && newyyv != 0)
				{
					yys = YYCOPY(newyys, yys, int);
					yyv = YYCOPY(newyyv, yyv, YYSTYPE);
				}
				else
					yynewmax = 0;	/* failed */
			}
			else				/* not first time */
			{
				yys = YYENLARGE(yys, int);
				yyv = YYENLARGE(yyv, YYSTYPE);
				if (yys == 0 || yyv == 0)
					yynewmax = 0;	/* failed */
			}
#endif
			if (yynewmax <= yymaxdepth)	/* tables not expanded */
			{
				yyerror( "yacc stack overflow" );
				YYABORT;
			}
			yymaxdepth = yynewmax;

			yy_ps = yys + yyps_index;
			yy_pv = yyv + yypv_index;
			yypvt = yyv + yypvt_index;
		}
		*yy_ps = yy_state;
		*++yy_pv = yyval;

		/*
		** we have a new state - find out what to do
		*/
	yy_newstate:
		if ( ( yy_n = yypact[ yy_state ] ) <= YYFLAG )
			goto yydefault;		/* simple state */
#if YYDEBUG
		/*
		** if debugging, need to mark whether new token grabbed
		*/
		yytmp = yychar < 0;
#endif
		if ( ( yychar < 0 ) && ( ( yychar = YYLEX() ) < 0 ) )
			yychar = 0;		/* reached EOF */
#if YYDEBUG
		if ( yydebug && yytmp )
		{
			register int yy_i;

			printf( "Received token " );
			if ( yychar == 0 )
				printf( "end-of-file\n" );
			else if ( yychar < 0 )
				printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ( ( yy_n += yychar ) < 0 ) || ( yy_n >= YYLAST ) )
			goto yydefault;
		if ( yychk[ yy_n = yyact[ yy_n ] ] == yychar )	/*valid shift*/
		{
			yychar = -1;
			yyval = yylval;
			yy_state = yy_n;
			if ( yyerrflag > 0 )
				yyerrflag--;
			goto yy_stack;
		}

	yydefault:
		if ( ( yy_n = yydef[ yy_state ] ) == -2 )
		{
#if YYDEBUG
			yytmp = yychar < 0;
#endif
			if ( ( yychar < 0 ) && ( ( yychar = YYLEX() ) < 0 ) )
				yychar = 0;		/* reached EOF */
#if YYDEBUG
			if ( yydebug && yytmp )
			{
				register int yy_i;

				printf( "Received token " );
				if ( yychar == 0 )
					printf( "end-of-file\n" );
				else if ( yychar < 0 )
					printf( "-none-\n" );
				else
				{
					for ( yy_i = 0;
						yytoks[yy_i].t_val >= 0;
						yy_i++ )
					{
						if ( yytoks[yy_i].t_val
							== yychar )
						{
							break;
						}
					}
					printf( "%s\n", yytoks[yy_i].t_name );
				}
			}
#endif /* YYDEBUG */
			/*
			** look through exception table
			*/
			{
				register int *yyxi = yyexca;

				while ( ( *yyxi != -1 ) ||
					( yyxi[1] != yy_state ) )
				{
					yyxi += 2;
				}
				while ( ( *(yyxi += 2) >= 0 ) &&
					( *yyxi != yychar ) )
					;
				if ( ( yy_n = yyxi[1] ) < 0 )
					YYACCEPT;
			}
		}

		/*
		** check for syntax error
		*/
		if ( yy_n == 0 )	/* have an error */
		{
			/* no worry about speed here! */
			switch ( yyerrflag )
			{
			case 0:		/* new error */
				yyerror( "syntax error" );
				goto skip_init;
			yyerrlab:
				/*
				** get globals into registers.
				** we have a user generated syntax type error
				*/
				yy_pv = yypv;
				yy_ps = yyps;
				yy_state = yystate;
			skip_init:
				yynerrs++;
				/* FALLTHRU */
			case 1:
			case 2:		/* incompletely recovered error */
					/* try again... */
				yyerrflag = 3;
				/*
				** find state where "error" is a legal
				** shift action
				*/
				while ( yy_ps >= yys )
				{
					yy_n = yypact[ *yy_ps ] + YYERRCODE;
					if ( yy_n >= 0 && yy_n < YYLAST &&
						yychk[yyact[yy_n]] == YYERRCODE)					{
						/*
						** simulate shift of "error"
						*/
						yy_state = yyact[ yy_n ];
						goto yy_stack;
					}
					/*
					** current state has no shift on
					** "error", pop stack
					*/
#if YYDEBUG
#	define _POP_ "Error recovery pops state %d, uncovers state %d\n"
					if ( yydebug )
						printf( _POP_, *yy_ps,
							yy_ps[-1] );
#	undef _POP_
#endif
					yy_ps--;
					yy_pv--;
				}
				/*
				** there is no state on stack with "error" as
				** a valid shift.  give up.
				*/
				YYABORT;
			case 3:		/* no shift yet; eat a token */
#if YYDEBUG
				/*
				** if debugging, look up token in list of
				** pairs.  0 and negative shouldn't occur,
				** but since timing doesn't matter when
				** debugging, it doesn't hurt to leave the
				** tests here.
				*/
				if ( yydebug )
				{
					register int yy_i;

					printf( "Error recovery discards " );
					if ( yychar == 0 )
						printf( "token end-of-file\n" );
					else if ( yychar < 0 )
						printf( "token -none-\n" );
					else
					{
						for ( yy_i = 0;
							yytoks[yy_i].t_val >= 0;
							yy_i++ )
						{
							if ( yytoks[yy_i].t_val
								== yychar )
							{
								break;
							}
						}
						printf( "token %s\n",
							yytoks[yy_i].t_name );
					}
				}
#endif /* YYDEBUG */
				if ( yychar == 0 )	/* reached EOF. quit */
					YYABORT;
				yychar = -1;
				goto yy_newstate;
			}
		}/* end if ( yy_n == 0 ) */
		/*
		** reduction by production yy_n
		** put stack tops, etc. so things right after switch
		*/
#if YYDEBUG
		/*
		** if debugging, print the string that is the user's
		** specification of the reduction which is just about
		** to be done.
		*/
		if ( yydebug )
			printf( "Reduce by (%d) \"%s\"\n",
				yy_n, yyreds[ yy_n ] );
#endif
		yytmp = yy_n;			/* value to switch over */
		yypvt = yy_pv;			/* $vars top of value stack */
		/*
		** Look in goto table for next state
		** Sorry about using yy_state here as temporary
		** register variable, but why not, if it works...
		** If yyr2[ yy_n ] doesn't have the low order bit
		** set, then there is no action to be done for
		** this reduction.  So, no saving & unsaving of
		** registers done.  The only difference between the
		** code just after the if and the body of the if is
		** the goto yy_stack in the body.  This way the test
		** can be made before the choice of what to do is needed.
		*/
		{
			/* length of production doubled with extra bit */
			register int yy_len = yyr2[ yy_n ];

			if ( !( yy_len & 01 ) )
			{
				yy_len >>= 1;
				yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
				yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
					*( yy_ps -= yy_len ) + 1;
				if ( yy_state >= YYLAST ||
					yychk[ yy_state =
					yyact[ yy_state ] ] != -yy_n )
				{
					yy_state = yyact[ yypgo[ yy_n ] ];
				}
				goto yy_stack;
			}
			yy_len >>= 1;
			yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
			yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
				*( yy_ps -= yy_len ) + 1;
			if ( yy_state >= YYLAST ||
				yychk[ yy_state = yyact[ yy_state ] ] != -yy_n )
			{
				yy_state = yyact[ yypgo[ yy_n ] ];
			}
		}
					/* save until reenter driver code */
		yystate = yy_state;
		yyps = yy_ps;
		yypv = yy_pv;
	}
	/*
	** code supplied by user is placed in this switch
	*/
	switch( yytmp )
	{
		
case 1:
# line 128 "text-parser.y"
{} break;
case 2:
# line 130 "text-parser.y"
{ } break;
case 3:
# line 131 "text-parser.y"
{ } break;
case 4:
# line 134 "text-parser.y"
{
					     current_object = new_object (yypvt[-0].strval, nil);
					     record_new_object (current_object);
					     clear_isa_parents (current_object);
					     clear_instance_of (current_object);
					     if (categories)
						free_list (categories);
					     categories = nil;
					     add_instance_of (current_object,
								new_object (yypvt[-1].strval, nil), false);
					} break;
case 5:
# line 146 "text-parser.y"
{ yyval.strval = yypvt[-0].strval; } break;
case 6:
# line 147 "text-parser.y"
{ yyval.strval = yypvt[-0].strval; } break;
case 7:
# line 149 "text-parser.y"
{
						    rule_variables = nil;
						    num_rule_variables = 0;
						    last_rule_variable = nil;
						} break;
case 8:
# line 155 "text-parser.y"
{
						    rule_variables = nil;
						    num_rule_variables = 0;
						    last_rule_variable = nil;
						} break;
case 10:
# line 163 "text-parser.y"
{ new_rule (yypvt[-1].strval, yypvt[-0].listval); } break;
case 11:
# line 164 "text-parser.y"
{ new_goals (yypvt[-0].listval); } break;
case 12:
# line 166 "text-parser.y"
{ yyval.listval = cons (yypvt[-0].atomval, nil); } break;
case 13:
# line 167 "text-parser.y"
{ yyval.listval = cons (yypvt[-1].atomval, yypvt[-0].listval); } break;
case 14:
# line 169 "text-parser.y"
{ yyval.atomval = new_atom (0, yypvt[-1].strval, yypvt[-0].listval); } break;
case 15:
# line 171 "text-parser.y"
{ yyval.atomval = new_atom (1, yypvt[-2].strval, yypvt[-1].listval); } break;
case 16:
# line 173 "text-parser.y"
{ yyval.atomval = yypvt[-1].atomval; } break;
case 17:
# line 175 "text-parser.y"
{ yyval.listval = nil; } break;
case 18:
# line 176 "text-parser.y"
{ yyval.listval = cons (yypvt[-1].termval, yypvt[-0].listval); } break;
case 19:
# line 178 "text-parser.y"
{ yyval.termval = new_object_term (yypvt[-0].objval); } break;
case 20:
# line 179 "text-parser.y"
{ yyval.termval = new_list_term (yypvt[-1].listval); } break;
case 21:
# line 180 "text-parser.y"
{ yyval.termval = new_var_term (yypvt[-0].varval); } break;
case 22:
# line 183 "text-parser.y"
{ yyval.varval = new_simple_var (yypvt[-1].strval, yypvt[-0].listval, 1); } break;
case 23:
# line 184 "text-parser.y"
{ yyval.varval = new_simple_var (yypvt[-1].strval, yypvt[-0].listval, 0); } break;
case 24:
# line 186 "text-parser.y"
{ yyval.listval = nil; } break;
case 25:
# line 187 "text-parser.y"
{ yyval.listval = yypvt[-1].listval; } break;
case 26:
# line 189 "text-parser.y"
{ yyval.listval = nil; } break;
case 27:
# line 190 "text-parser.y"
{ yyval.listval = cons (yypvt[-1].objval, yypvt[-0].listval); } break;
case 28:
# line 194 "text-parser.y"
{
					     current_object = new_object (yypvt[-1].strval, nil);
					     record_new_object (current_object);
					     clear_isa_parents (current_object);
					     clear_instance_of (current_object);
					     if (categories)
						free_list (categories);
					     categories = nil;
					     add_instance_of (current_object,
								new_object (yypvt[-0].strval, nil), false);
					} break;
case 29:
# line 207 "text-parser.y"
{
					     current_object = new_object (yypvt[-1].strval, nil);
					     record_new_object (current_object);
					     clear_isa_parents (current_object);
					     clear_instance_of (current_object);
					     if (categories)
						free_list (categories);
					     categories = nil;
					     add_instance_of (current_object,
								new_object (yypvt[-0].strval, nil), false);
					} break;
case 30:
# line 220 "text-parser.y"
{ minimize_instances (current_object); } break;
case 33:
# line 227 "text-parser.y"
{ yyval.strval = TOKEN_KEYWORD; } break;
case 34:
# line 228 "text-parser.y"
{ yyval.strval = SCLASS_KEYWORD; } break;
case 35:
# line 229 "text-parser.y"
{ yyval.strval = M1CLASS_KEYWORD; } break;
case 36:
# line 230 "text-parser.y"
{ yyval.strval = M2CLASS_KEYWORD; } break;
case 37:
# line 232 "text-parser.y"
{ yyval.strval = yypvt[-0].strval; } break;
case 40:
# line 237 "text-parser.y"
{ add_instance_of (current_object, yypvt[-1].objval, false); } break;
case 41:
# line 239 "text-parser.y"
{ minimize_instances (current_object); } break;
case 43:
# line 242 "text-parser.y"
{ add_isa_parent (current_object, yypvt[-1].objval, false); } break;
case 45:
# line 246 "text-parser.y"
{ add_instance_of (current_object, yypvt[-1].objval, false); } break;
case 47:
# line 250 "text-parser.y"
{} break;
case 48:
# line 251 "text-parser.y"
{ add_instance_of (current_object, yypvt[-1].objval, false); } break;
case 50:
# line 254 "text-parser.y"
{ add_isa_parent (current_object, yypvt[-1].objval, false); } break;
case 52:
# line 258 "text-parser.y"
{} break;
case 53:
# line 259 "text-parser.y"
{ add_isa_parent (current_object, yypvt[-1].objval, false); } break;
case 54:
# line 261 "text-parser.y"
{
					if (categories)
					    free_list (categories);
					categories = resolve_cat (current_object, yypvt[-0].strval, nil);
				  } break;
case 55:
# line 267 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].strval, nil); } break;
case 56:
# line 268 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].strval, nil); } break;
case 57:
# line 269 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].strval, nil); } break;
case 59:
# line 273 "text-parser.y"
{
			   create_attr (current_object, nil, categories, yypvt[-1].objval);
			} break;
case 67:
# line 288 "text-parser.y"
{
					    categories = resolve_cat (current_object, yypvt[-1].strval, nil);
					 } break;
case 69:
# line 293 "text-parser.y"
{
					    if (categories)
						free_list (categories);
					    categories = resolve_cat (current_object, yypvt[-1].strval, nil);
					 } break;
case 70:
# line 299 "text-parser.y"
{ yyval.listval = nil; } break;
case 71:
# line 300 "text-parser.y"
{ yyval.listval = cons (yypvt[-1].objval, yypvt[-0].listval); } break;
case 72:
# line 301 "text-parser.y"
{ yyval.listval = cons (yypvt[-2].objval, yypvt[-0].listval); } break;
case 73:
# line 303 "text-parser.y"
{ yyval.listval = yypvt[-1].listval; } break;
case 74:
# line 306 "text-parser.y"
{
						current_thing = yypvt[-0].strval;
						current_sequence = nil;
					    } break;
case 75:
# line 310 "text-parser.y"
{
						current_thing = nil;
						current_sequence = yypvt[-0].listval;
					    } break;
case 76:
# line 316 "text-parser.y"
{ attribute_from_current (); } break;
case 77:
# line 317 "text-parser.y"
{ attribute_from_object (yypvt[-0].objval); } break;
case 78:
# line 318 "text-parser.y"
{ attribute_from_sequence (yypvt[-0].listval); } break;
case 81:
# line 324 "text-parser.y"
{
		    } break;
case 86:
# line 334 "text-parser.y"
{} break;
case 87:
# line 336 "text-parser.y"
{ } break;
case 88:
# line 338 "text-parser.y"
{} break;
case 89:
# line 339 "text-parser.y"
{ } break;
case 91:
# line 343 "text-parser.y"
{
					    if (categories)
						free_list (categories);
					    categories = resolve_cat (current_object, yypvt[-1].strval, nil);
					 } break;
case 92:
# line 349 "text-parser.y"
{} break;
case 93:
# line 351 "text-parser.y"
{
			    categories = resolve_cat (current_object, yypvt[-1].strval, categories);
			} break;
case 94:
# line 355 "text-parser.y"
{ yyval.objval = yypvt[-0].objval; } break;
case 95:
# line 356 "text-parser.y"
{ yyval.objval = 0; } break;
case 96:
# line 360 "text-parser.y"
{
		    if (yypvt[-1].objval)
			create_attr (current_object, yypvt[-3].strval, categories, yypvt[-1].objval);
		} break;
case 97:
# line 365 "text-parser.y"
{
		    create_attr (current_object, nil, categories, yypvt[-1].objval);
		} break;
case 98:
# line 370 "text-parser.y"
{} break;
case 99:
# line 371 "text-parser.y"
{} break;
case 100:
# line 374 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].strval, nil); } break;
case 101:
# line 375 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].strval, nil); } break;
case 102:
# line 376 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].strval, nil); } break;
case 103:
# line 377 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].strval, nil); } break;
case 104:
# line 378 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].sval, nil); } break;
case 105:
# line 379 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].sval, nil); } break;
case 106:
# line 380 "text-parser.y"
{ yyval.objval = new_object (yypvt[-0].sval, nil); } break;
case 107:
# line 383 "text-parser.y"
{ yyval.strval = yypvt[-0].strval; } break;
case 108:
# line 384 "text-parser.y"
{ yyval.strval = yypvt[-0].strval; } break;
case 109:
# line 385 "text-parser.y"
{ yyval.strval = yypvt[-0].strval; } break;
case 110:
# line 386 "text-parser.y"
{ yyval.strval = yypvt[-0].strval; } break;
case 111:
# line 387 "text-parser.y"
{ yyval.strval = yypvt[-0].sval; } break;
case 112:
# line 388 "text-parser.y"
{ yyval.strval = yypvt[-0].sval; } break;
case 113:
# line 389 "text-parser.y"
{ yyval.strval = yypvt[-0].sval; } break;
case 114:
# line 392 "text-parser.y"
{} break;
	}
	goto yystack;		/* reset registers in driver code */
}

