%{
/*
 * text-parser.y --
 *
 *  Last Modified: Mon Sep 18 10:02:13 1995 by Bryan M. Kramer
 *
 * Copyright (c) 1990-1994 University of Toronto.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF Toronto BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * Toronto HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF Toronto SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF Toronto HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

# include <stdio.h>
# include <stdlib.h>
# include "kb.h"
# include "dr.h"

# define yywrap() 1

typedef union {
    int ival;
    float fval;
    void *sval;
    char* strval;
    KBObject* objval;
    KBList* listval;
    Atom*   atomval;
    Term*   termval;
    Variable* varval;
} YYSTYPE;

KBObject* current_object;
KBList*  categories;
KBList*	 current_sequence;
char*    current_thing;

extern VarList*    rule_variables;
extern int	   num_rule_variables;
extern VarList*    last_rule_variable;


void attribute_from_current ();
void attribute_from_object ();
void attribute_from_sequence ();

# define TOKEN_KEYWORD "Token"
# define SCLASS_KEYWORD "SClass"
# define M1CLASS_KEYWORD "M1Class"
# define M2CLASS_KEYWORD "M2Class"

#undef YYMAXDEPTH
#define YYMAXDEPTH 10000
#undef YYERROR_VERBOSE
%}


%start file

%token <sval> TP_RULE
%token <sval> TP_NOT
%token <sval> TP_GOAL
%token <sval> TP_TOKEN
%token <ival> TP_SCLASS
%token <ival> TP_M1CLASS
%token <ival> TP_M2CLASS
%token <ival> TP_M3CLASS
%token <ival> TP_IN
%token <ival> TP_ISA
%token <ival> TP_WITH
%token <ival> TP_END
%token <ival> TP_COMMA
%token <ival> TP_COLON
%token <ival> TP_SLASH
%token <ival> TP_DOLLAR
%token <ival> TP_EQUAL
%token <ival> TP_nil
%token <ival> TP_SEMICOLON
%token <ival> INSTANCE
%token <ival> SLOT
%token <ival> LBRACKET
%token <ival> RBRACKET
%token <ival> LBRACE
%token <ival> RBRACE
%token <ival> START_SEQ
%token <ival> END_SEQ
%token <strval> TP_IDENTIFIER
%token <strval> TP_STRING
%token <strval> TP_OBJECT

%token <sval> TP_DASH

%type	<strval> levelKeyword
%type	<strval> cname
%type	<strval> name
%type   <objval> ObjectID
%type   <objval> optObjectID
%type   <objval> instanceSlotValueValue
%type   <strval> thing
%type   <listval> objects
%type   <listval> sequence
%type	<listval> atoms
%type	<listval> terms
%type	<listval> var_types
%type	<listval> term_objects
%type	<atomval> one_atom
%type	<atomval> atom_kind
%type	<termval> one_term
%type	<varval>  variable

%%
/* Rules */

file	     : declarations  {}

declarations : /* empty */ { }
	     | declaration declarations { }


objDecl	     : levelKeyword name	{
					     current_object = new_object ($2, nil);
					     record_new_object (current_object);
					     clear_isa_parents (current_object);
					     clear_instance_of (current_object);
					     if (categories)
						free_list (categories);
					     categories = nil;
					     add_instance_of (current_object,
								new_object ($1, nil), false);
					}

cname	    :	name	      { $$ = $1; }
	    |	levelKeyword  { $$ = $1; }

goalStart   :	TP_GOAL				{
						    rule_variables = nil;
						    num_rule_variables = 0;
						    last_rule_variable = nil;
						}

ruleStart   :	TP_RULE				{
						    rule_variables = nil;
						    num_rule_variables = 0;
						    last_rule_variable = nil;
						}


listDecl    :	instanceHead instanceList isaList withList
	    |	ruleStart name atoms		{ new_rule ($2, $3); }
	    |	goalStart atoms			{ new_goals ($2); }

atoms	    :	one_atom			{ $$ = cons ($1, nil); }
	    |	one_atom atoms			{ $$ = cons ($1, $2); }

atom_kind   :	name terms			{ $$ = new_atom (0, $1, $2); }
	    |	TP_NOT LBRACKET name terms RBRACKET
						{ $$ = new_atom (1, $3, $4); }

one_atom    :	LBRACKET atom_kind RBRACKET	{ $$ = $2; }

terms	    :	/* empty */			{ $$ = nil; }
	    |	one_term terms			{ $$ = cons ($1, $2); }

one_term    :	ObjectID			{ $$ = new_object_term ($1); }
	    |	LBRACE term_objects RBRACE	{ $$ = new_list_term ($2); }
	    |	variable			{ $$ = new_var_term ($1); }


variable    :	TP_EQUAL name var_types		{ $$ = new_simple_var ($2, $3, 1); }
	    |	TP_DOLLAR name var_types	{ $$ = new_simple_var ($2, $3, 0); }

var_types   :	/* empty */			{ $$ = nil; }
	    |	TP_SLASH term_objects TP_SLASH	{ $$ = $2; }

term_objects  :	/* empty */			{ $$ = nil; }
	    |	ObjectID term_objects		{ $$ = cons ($1, $2); }


instanceHead    :	cname cname
					{
					     current_object = new_object ($1, nil);
					     record_new_object (current_object);
					     clear_isa_parents (current_object);
					     clear_instance_of (current_object);
					     if (categories)
						free_list (categories);
					     categories = nil;
					     add_instance_of (current_object,
								new_object ($2, nil), false);
					}

instanceDecl     : INSTANCE name cname
					{
					     current_object = new_object ($2, nil);
					     record_new_object (current_object);
					     clear_isa_parents (current_object);
					     clear_instance_of (current_object);
					     if (categories)
						free_list (categories);
					     categories = nil;
					     add_instance_of (current_object,
								new_object ($3, nil), false);
					}				

declaration  : objDecl inClause isaClause withClause endKeyword
			{ minimize_instances (current_object); }
	     | LBRACKET listDecl RBRACKET
	     | instanceDecl instanceSlots




levelKeyword : TP_TOKEN { $$ = TOKEN_KEYWORD; }
             | TP_SCLASS { $$ = SCLASS_KEYWORD; }
             | TP_M1CLASS { $$ = M1CLASS_KEYWORD; }
             | TP_M2CLASS { $$ = M2CLASS_KEYWORD; }

name         : TP_IDENTIFIER { $$ = $1; }

instanceList	: LBRACKET instanceAtom RBRACKET

instanceAtom	: /* empty */
		| ObjectID instanceAtom { add_instance_of (current_object, $1, false); }

isaList		: LBRACKET isaAtom RBRACKET { minimize_instances (current_object); }

isaAtom		: /* empty */
		| ObjectID isaAtom { add_isa_parent (current_object, $1, false); }


inClause       : /* empty */
             | inKeyword ObjectID inClauseIDs { add_instance_of (current_object, $2, false); }

inKeyword    : TP_IN

inClauseIDs  : /* empty */ {}
             | TP_COMMA ObjectID inClauseIDs { add_instance_of (current_object, $2, false); }

isaClause    : /* empty */
             | isaKeyword ObjectID isaClauseIDs { add_isa_parent (current_object, $2, false); }

isaKeyword   : TP_ISA

isaClauseIDs   : /* empty */ {}
             | TP_COMMA ObjectID isaClauseIDs { add_isa_parent (current_object, $2, false); }

instanceSlotStart :	SLOT name {
					if (categories)
					    free_list (categories);
					categories = resolve_cat (current_object, $2, nil);
				  }

instanceSlotValueValue	:   TP_IDENTIFIER   { $$ = new_object ($1, nil); }
			|   TP_STRING	    { $$ = new_object ($1, nil); }
			|   TP_OBJECT	    { $$ = new_object ($1, nil); }

instanceSlotValues  :	/* empty */
		    |	instanceSlotValueValue instanceSlotValues
			{
			   create_attr (current_object, nil, categories, $1);
			}

instanceSlots	:    /* empty */
		|   instanceSlotStart instanceSlotValues instanceSlots

withList	:   LBRACKET withGroup RBRACKET

withGroup	:   /* empty */
		|   LBRACKET withGroupCats withGroupAttrs RBRACKET withGroup

withGroupCats	:   LBRACKET withGroupCats1 RBRACKET

withGroupCats2	:   /* empty */
		|   name withGroupCats2  {
					    categories = resolve_cat (current_object, $1, nil);
					 }

withGroupCats1	:   /* empty */
		|   name withGroupCats2  {
					    if (categories)
						free_list (categories);
					    categories = resolve_cat (current_object, $1, nil);
					 }

objects		:   /* empty */			{ $$ = nil; }
		|   ObjectID objects		{ $$ = cons ($1, $2); }
		|   LBRACKET ObjectID RBRACKET objects	{ $$ = cons ($2, $4); }

sequence	:   START_SEQ objects END_SEQ   { $$ = $2; }				


withGroupAttrs2Head	:	thing	    {
						current_thing = $1;
						current_sequence = nil;
					    }
			|	sequence    {
						current_thing = nil;
						current_sequence = $1;
					    }


withGroupDoIt		:   /* empty */	    { attribute_from_current (); }
			|   ObjectID	    { attribute_from_object ($1); }
			|   sequence	    { attribute_from_sequence ($1); }

withGroupAttrs2Tail	:   withGroupDoIt RBRACKET withGroupAttrs2

withGroupAttrs2	:   /* empty */
		|   LBRACKET  withGroupAttrs2Head withGroupAttrs2Tail
		    {
		    }

withGroupAttrs	:   /* empty */
		|   LBRACKET withGroupAttrs2 RBRACKET
			

withClause   : /* empty */
             | withKeyword attrDecls

withKeyword  : TP_WITH {}

attrDecls    : attrDecls2 { }

attrDecls2   : /* empty */ {}
             | attrDecl attrDecls2 { }

attrDecl     : categories valuePairs

categories   : TP_IDENTIFIER categories2 {
					    if (categories)
						free_list (categories);
					    categories = resolve_cat (current_object, $1, nil);
					 }

categories2  : /* empty */ {}
             | TP_COMMA TP_IDENTIFIER categories2
			{
			    categories = resolve_cat (current_object, $2, categories);
			}

optObjectID : ObjectID			 { $$ = $1; }
	    | TP_nil			 { $$ = 0; }


valuePairs   : TP_IDENTIFIER TP_COLON optObjectID valuePairs2
		{
		    if ($3)
			create_attr (current_object, $1, categories, $3);
		}
             | TP_COLON ObjectID valuePairs2
		{
		    create_attr (current_object, nil, categories, $2);
		}


valuePairs2  :  /* empty */  {}
 	     | TP_SEMICOLON valuePairs {}


ObjectID    :	TP_IDENTIFIER     { $$ = new_object ($1, nil); }
	    |	levelKeyword      { $$ = new_object ($1, nil); }
	    |	TP_OBJECT         { $$ = new_object ($1, nil); }
	    |	TP_STRING         { $$ = new_object ($1, nil); }
	    |	TP_RULE		  { $$ = new_object ($1, nil); }
	    |	TP_GOAL		  { $$ = new_object ($1, nil); }
	    |	TP_NOT		  { $$ = new_object ($1, nil); }


thing	    :	TP_IDENTIFIER     { $$ = $1; }
	    |	levelKeyword      { $$ = $1; }
	    |	TP_OBJECT         { $$ = $1; }
	    |	TP_STRING         { $$ = $1; }
	    |	TP_RULE           { $$ = $1; }
	    |	TP_GOAL           { $$ = $1; }
	    |	TP_NOT            { $$ = $1; }


endKeyword  : TP_END {}

%%
/* Programs */


int use_string;
int string_pos;
int string_len;
char* input_string;


#include "scanner.h"


int parse_error;
char* cur_file;

void yyerror (s)
    char	*s;
{
    parse_error = 1;
    if (use_string)
	fprintf (stderr, "%s at position %d in string '%s' near %s\n", s, string_pos,
		 input_string, yytext);
    else
	fprintf(stderr, "Syntax error on line %d of %s near %s\n",
		yylineno, cur_file, yytext);
}


void semantic_msg (s)
    char	*s;
{
    parse_error = 1;
    if (use_string)
	fprintf (stderr, "%s at %d in string '%s' near %s\n",
		 s, string_pos, input_string, yytext);
    else
	fprintf (stderr, "%s on line %d of %s near %s\n", s, yylineno, cur_file, yytext);
}



int parse_string (str)
    char* str;
{
    use_string = 1;
    string_pos = 0;
    string_len = strlen (str);
    input_string = str;
    reset_new_object_list ();
    yyparse ();
    return parse_error;
}



int parse_file (file)
    char* file;
{
    Source* save_source;
    save_source = current_source;
    current_source = new (Source);
    current_source->source_type = SOURCE_FILE;
    current_source->source.file.name = strdup (file);

    cur_file = file;
    use_string = 0;
    current_object = nil;
    categories = nil;
    parse_error = 0;
    yylineno = 1;
    yyin = fopen (cur_file, "r");
    if (! yyin)
    {
	fprintf (stderr, "Couldn't open file %s\n", cur_file);
	return;
    }
    reset_new_object_list ();
    yyparse ();

    fclose (yyin);

    current_source = save_source;

    return parse_error;
}


/* current thing and current sequence are mutually exclusive */

void attribute_from_current ()
{
    KBList* ptr;
    if (current_thing)
    {
	create_attr (current_object, nil, categories, new_object (current_thing, nil));
    }
    else
    {
	for (ptr = current_sequence; ptr; ptr = ptr->next)
	    create_attr (current_object, nil, categories, ptr->object);

	free_list (current_sequence);
    }
}


void attribute_from_object (object)
    KBObject* object;
{
    KBList* ptr;
    if (current_thing)
    {
	create_attr (current_object, current_thing, categories, object);
    }
    else
    {
	for (ptr = current_sequence; ptr; ptr = ptr->next)
	    create_attr (current_object, ptr->object->name, categories, object);

	free_list (current_sequence);
    }
}



void attribute_from_sequence (seq)
    KBList* seq;
{
    KBList* ptr;
    KBList* ptr2;
    if (current_thing)
    {
	for (ptr = seq; ptr; ptr = ptr->next)
	    create_attr (current_object, current_thing, categories, ptr->object);
    }
    else
    {
	for (ptr = seq; ptr; ptr = ptr->next)
	{
	    for (ptr2 = current_sequence; ptr; ptr = ptr->next)
		create_attr (current_object, ptr->object->name, categories, ptr2->object);
	}

	free_list (current_sequence);
    }

    free_list (seq);
}
