/* 
 * tkAppInit.c --
 *
 *	Provides a default version of the Tcl_AppInit procedure for
 *	use in wish and similar Tk-based applications.
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*         Copyright (c) 1995  University of Toronto, Ontario Hydro           */
/* ======================================================================     */
/* Permission to use, copy, modify, and distribute this software and its      */
/* documentation for any non-commercial purpose is hereby granted,            */
/* provided that the above copyright notice appear in all copies and that     */
/* both that the copyright notice and warranty disclaimer appear in           */
/* supporting documentation, and that the names of the University of Toronto, */
/* Ontario Hydro, and Bryan M. Kramer not be used in advertising or publicity */
/* pertaining to distribution of the software without specific, written prior */
/* permission.                        					      */
/*                                                                            */
/* The University of Toronto, Ontario Hydro, and Bryan M. Kramer disclaim all */
/* warranties with regard to this software, including all implied warranties  */
/* of merchantability and fitness.  In no event shall the University of Toronto, */
/* Ontario Hydro, or Bryan M. Kramer be liable for any special, indirect or   */
/* consequential damages or any damages whatsoever resulting from loss of use, */
/* data or profits, whether in an action of contract, negligence or other     */
/* tortuous action, arising out of or in connection with the use or           */
/* performance of this software.                                              */
/*                                                                            */
#ifndef lint
static char sccsid[] = "@(#) tkAppInit.c 1.12 94/12/17 16:30:56";
#endif /* not lint */

#include "tk.h"



# ifdef TKWWW
# include <TkWWWCmds.h>
# include <TkWWWBitmap.h>
# endif

/*
 * The following variable is a special hack that is needed in order for
 * Sun shared libraries to be used for Tcl.
 */

#ifdef NEED_MATHERR
extern int matherr();
int *tclDummyMathPtr = (int *) matherr;
#endif

/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *	This is the main program for the application.
 *
 * Results:
 *	None: Tk_Main never returns here, so this procedure never
 *	returns either.
 *
 * Side effects:
 *	Whatever the application does.
 *
 *----------------------------------------------------------------------
 */

int
main(argc, argv)
    int argc;			/* Number of command-line arguments. */
    char **argv;		/* Values of command-line arguments. */
{
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;			/* Needed only to prevent compiler warning. */
}


Tcl_Interp* global_interp;



/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

int apacs_www;

int
Tcl_AppInit(interp)
    Tcl_Interp *interp;		/* Interpreter for application. */
{
    Tk_Window main;

    main = Tk_MainWindow(interp);

    global_interp = interp;

    if (Tcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * Call the init procedures for included packages.  Each call should
     * look like this:
     *
     * if (Mod_Init(interp) == TCL_ERROR) {
     *     return TCL_ERROR;
     * }
     *
     * where "Mod" is the name of the module.
     */

/*    if (Blt_Init(interp) == TCL_ERROR) { */
/*	return TCL_ERROR; */
/*    } */

# ifdef TKWWW
    if (WWW_AppInit(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

    if (WWW_BitmapInit(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    apacs_www = 1;
# else
    apacs_www = 0;
# endif

    /*
     * Call Tcl_CreateCommand for application-specific commands, if
     * they weren't already created by the init procedures called above.
     */

    registerKBCommands (interp);
    Tcl_LinkVar (interp, "apacs_www", (char*) &apacs_www, TCL_LINK_INT);

    /*
     * Specify a user-specific startup file to invoke if the application
     * is run interactively.  Typically the startup file is "~/.apprc"
     * where "app" is the name of the application.  If this line is deleted
     * then no user-specific startup file will be run under any conditions.
     */
#if (TCL_MAJOR_VERSION >= 7) && (TCL_MINOR_VERSION >= 5)
    Tcl_SetVar(interp, "tcl_rcFileName", "~/.wishrc", TCL_GLOBAL_ONLY);
#else
    tcl_RcFileName = "~/.wishrc";
#endif
    return TCL_OK;
}
